import logging
import re

import pandas as pd

from langchain.output_parsers import ResponseSchema
from langchain_core.prompts import PromptTemplate
from tqdm import tqdm

from intugle.core.llms.chat import ChatModelLLM
from intugle.core.settings import settings

log = logging.getLogger(__name__)


class KeyIdentificationLLM:
    KI_PROMPT_TEMPLATE = template = """Identify ONLY SINGLE primary key of a database table using profiling metadata of the schema.

- Review the provided metadata to determine which column is most likely the primary key in a database schema.
- The primary key should uniquely identify each record and usually has a unique constraint and no null values.

# Steps

1. **Examine Uniqueness**: Identify column that have a unique constraint or a high percentage of unique values compared to the total number of records.
2. **Check for Null Values**: Ensure the column considered for the primary key do not allow null values.
3. **DONOT Identify Combinations**: If no single column meets the criteria, then DONOT  identify multiple columns as composite key.
4. DONOT Identify Index Keys as a primary key ( They are usually integer type with, sequence of values that are increasing by 1 )
5. created_date, modified_date, updated_date should not be considered as primary key.

PROFILING_METADATA: {profiling_data}

INSTRUCTIONS: {format_instructions}"""

    primary_key = [ResponseSchema(name='PRIMARY KEY', description="Returns ONLY a SINGLE potential primary key."),]
    LLM_CONFIG = {
        "temperature": 1,
        # "model_kwargs":{
        "top_p": 0.5,
        "frequency_penalty": 0,
        "presence_penalty": 0,
        "stop": "END",
        # }
    }

    def __init__(self, profiling_data: pd.DataFrame,
                 llm: ChatModelLLM = None,
                 *args, **kwargs):

        self.__chat_llm = ChatModelLLM.build(
            model_name=settings.LLM_PROVIDER,
            llm_config=self.LLM_CONFIG,
            prompt_template=PromptTemplate,
            template_string=self.KI_PROMPT_TEMPLATE,
            response_schemas=self.primary_key,
     

        ) if llm is None else llm

        self.profiling_data = profiling_data.loc[
            (profiling_data["uniqueness"] >= 0.8) & (profiling_data["datatype_l2"] == "dimension")
        ].reset_index(drop=True)[["table_name", "column_name", "count", "distinct_count", "null_count", "datatype_l1", "sample_data"]]

    @classmethod
    def fetch_primary_key(cls, llm: ChatModelLLM, profiling_data, table_name):
        response, parsing_success, prompt = llm.invoke(
            profiling_data=profiling_data,
            metadata={"table": table_name}
        )
        
        if parsing_success:
            return response["PRIMARY KEY"], prompt
        else:
            try:
                pattern = r'"PRIMARY KEY":\s*"(\w+)"'
                match = re.search(pattern, str(response))
                return match.group(1), prompt
            except Exception as ex:
                log.warning(f"[!] Error while parsing : {ex}")
                return "", prompt
                    
    def __format_profiling_data__(self, profiling_data: pd.DataFrame, table_name: str) -> str:
        
        profiling_txt = f"Table Name: {table_name}\n\n"

        temp = profiling_data[["column_name", "count", "distinct_count", "null_count", "datatype_l1", "sample_data"]]

        temp.rename(columns={"datatype_l1": "datatype"}, inplace=True)

        if temp.shape[0] == 0:

            return None
        
        profiling_txt += str(temp.to_dict(orient="records"))

        return profiling_txt
    
    def __run_ki__(self, table_name: str, prof_data: pd.DataFrame):

        profiling_txt = self.__format_profiling_data__(profiling_data=prof_data, table_name=table_name)

        primary_key = ""

        prompt = ""

        if prof_data is not None:
            
            primary_key, prompt = self.fetch_primary_key(llm=self.__chat_llm, profiling_data=profiling_txt, table_name=table_name)
            
            if primary_key not in prof_data["column_name"].values.tolist():
                
                log.warning(f"[!] {primary_key} key identified by LLM is not a valid column of {table_name} table")
                
                primary_key = ""

        return (table_name, prompt, primary_key)
    
    def __post_process__(self, ki_llm_result: pd.DataFrame) -> pd.DataFrame:
                
        ki_llm_result.fillna("", inplace=True)
        
        ki_llm_result = ki_llm_result[ki_llm_result.predicted_key_llm != ""].reset_index(drop=True)
        
        if (ki_llm_result.shape[0] <= 0):
            return None
        
        ki_llm_result["column_name"] = ki_llm_result["predicted_key_llm"]
        
        ki_llm_result["predicted_key_llm"] = 1
        
        ki_llm_result["predicted_key_prob_llm"] = settings.KI_CONFIG["RULES"]["LLM_PROBABILITY"]
        
        return ki_llm_result
    
    def __call__(self):
        
        ki_llm_result = pd.DataFrame()

        if self.profiling_data.shape[0] == 0:
            
            return ki_llm_result
        
        results = []

        pb = tqdm(self.profiling_data.groupby("table_name"))

        for (table_name), prof_data in pb:
            
            pb.set_description(desc=f"[*] {table_name}")
            
            pk_info = self.__run_ki__(table_name, prof_data)
            
            results.append(pk_info)
        
        ki_llm_result = pd.DataFrame(results, columns=["table_name", "prompt", "predicted_key_llm"])
        
        ki_llm_result = self.__post_process__(ki_llm_result=ki_llm_result)
        
        if ki_llm_result is not None and not ki_llm_result.empty:
            return ki_llm_result.iloc[0].to_dict()
        return {}



from typing import List, Optional

from pydantic import Field

from intugle.common.resources.base import BaseResource
from intugle.common.schema import NodeType, SchemaBase
from intugle.models.resources.model import Column, ModelProfilingMetrics


class SourceTables(SchemaBase):
    name: str
    description: str
    tags: Optional[List[str]] = Field(default_factory=list)
    # details: dict
    columns: List[Column] = Field(default_factory=list)
    profiling_metrics: Optional[ModelProfilingMetrics] = None


class Source(BaseResource):
    schema: str
    database: str
    resource_type: NodeType = NodeType.SOURCE
    table: SourceTables = Field(default_factory=list)



class Transformations:
    # Aggregate Functions
    @classmethod
    def aggregate(cls, expr):
        return expr

    # Type Conversions
    @classmethod
    def toInt(cls, colm):
        return "int({})".format(colm)

    @classmethod
    def toBigInt(cls, colm):
        return "bigint({})".format(colm)

    @classmethod
    def toFloat(cls, colm):
        return "float({})".format(colm)

    @classmethod
    def toDouble(cls, colm):
        return "double({})".format(colm)

    @classmethod
    def toDate(cls, colm):
        return "date({})".format(colm)

    @classmethod
    def toBoolean(cls, colm):
        return "boolean({})".format(colm)

    @classmethod
    def toString(cls, colm):
        return "string({})".format(colm)

    # String Functions
    @classmethod
    def concat(cls, colmns: list):
        return "concat({})".format(",".join(colmns))

    @classmethod
    def concat_ws(cls, sep: str, colmns: list):
        return "concat_ws('{}',{})".format(sep, ",".join(colmns))

    @classmethod
    def trim(cls, colm):
        return "trim({})".format(colm)

    @classmethod
    def ltrim(cls, colm):
        return "ltrim({})".format(colm)

    @classmethod
    def rtrim(cls, colm):
        return "rtrim({})".format(colm)

    @classmethod
    def upper(cls, colm):
        return "upper({})".format(colm)

    @classmethod
    def capitalize(cls, colm):
        return "initcap({})".format(colm)

    @classmethod
    def lower(cls, colm):
        return "lower({})".format(colm)

    @classmethod
    def length(cls, colm):
        return "length({})".format(colm)

    @classmethod
    def right(cls, colm, length: int):
        return "right({}, {})".format(colm, length)

    @classmethod
    def left(cls, colm, length: int):
        return "left({}, {})".format(colm, length)

    @classmethod
    def coalesce(cls, exprs: list):
        return "coalesce({})".format(",".join(exprs))

    @classmethod
    def replace(cls, colm, search: str, replace: str = ""):
        if replace == "":
            return "replace({}, {})".format(colm, search)
        return "replace({}, {}, {})".format(colm, search, replace)

    @classmethod
    def reverse(cls, colm):
        return "reverse({})".format(colm)

    # Numeric Functions

    @classmethod
    def absolute(cls, colm):
        return "abs({})".format(colm)

    @classmethod
    def average(cls, expr):
        return "avg({})".format(expr)

    @classmethod
    def ceil(cls, colm):
        return "ceil({})".format(colm)

    @classmethod
    def floor(cls, colm):
        return "floor({})".format(colm)

    @classmethod
    def maximum(cls, colm):
        return "max({})".format(colm)

    @classmethod
    def mean(cls, colm):
        return "mean({})".format(colm)

    @classmethod
    def minimum(cls, colm):
        return "min({})".format(colm)

    @classmethod
    def modulus(cls, expr1, expr2):
        return "mod({}, {})".format(expr1, expr2)

    @classmethod
    def negative(cls, expr):
        return "negative({})".format(expr)

    @classmethod
    def power(cls, expr1, expr2):
        return "power({}, {})".format(expr1, expr2)

    @classmethod
    def round(cls, expr, decimal):
        return "round({}, {})".format(expr, decimal)

    @classmethod
    def square_root(cls, expr):
        return "sqrt({})".format(expr)

    @classmethod
    def stddev(cls, expr):
        return "stddev({})".format(expr)

    @classmethod
    def sum(cls, expr):
        return "sum({})".format(expr)

    @classmethod
    def variance(cls, expr):
        return "variance({})".format(expr)

    @classmethod
    def count(cls, expr):
        return "count({})".format(expr)

    @classmethod
    def countDistinct(cls, expr):
        return "count(distinct {})".format(expr)

    @classmethod
    def currentDate(cls):
        return "current_date()"

    @classmethod
    def day(cls, expr):
        return "day({})".format(expr)

    @classmethod
    def month(cls, expr):
        return "month({})".format(expr)

    @classmethod
    def year(cls, expr):
        return "year({})".format(expr)

    @classmethod
    def yearMonth(cls, expr):
        return "date_format({}, 'yyyy-MM')".format(expr)

    @classmethod
    def currentTimestamp(cls, expr):
        return "current_timestamp({})".format(expr)

    @classmethod
    def dateFormat(cls, expr, dt_format):
        return "date_format({}, '{}')".format(expr, dt_format)

    # Modifier Functions

    @classmethod
    def distinct(cls, expr):
        return "distinct {}".format(expr)

    @classmethod
    def custom(cls, expr):
        return "{}".format(expr)



import logging

from typing import List, Optional, TypedDict

import pandas as pd

from pydantic import BaseModel, Field

from intugle.core import settings
from intugle.core.utilities.processing import classify_datetime_format, preprocess_profiling_data

log = logging.getLogger(__name__)


class linkage(BaseModel):
    table1: str = Field(description="Verbatim name of table1 or NA")
    column1: str = Field(description="Verbatim name of column1 in table1 or NA")
    table2: str = Field(description="Verbatim name of table2 or NA")
    column2: str = Field(description="Verbatim name of column2 in table2 or NA")


def extract_innermost_dict(d):
    if isinstance(d, dict):
        for key, value in d.items():
            if isinstance(value, dict):
                # Recursively search for the innermost dictionary
                return extract_innermost_dict(value)
    return d  # Return the current dictionary when no further dictionaries are found


FEASIBLE_DTYPES = {
    "group1": ["integer", "float"],
    "group2": ["alphanumeric", "close_ended_text"],
    "group3": ["close_ended_text", "open_ended_text"]
}


class GraphState(TypedDict):
    input_text: str
    potential_link: dict
    error_msg: List[str]
    iteration: int
    link_type: str
    if_error: bool
    intersect_count: Optional[int] = None
    intersect_ratio_col1: Optional[float] = None
    intersect_ratio_col2: Optional[float] = None
    accuracy: Optional[float] = None


def dtype_check(dtype1: str, dtype2: str) -> bool:

    if dtype1 == dtype2:
        return True
    
    for _, feasible_dtypes in FEASIBLE_DTYPES.items():

        if dtype1 in feasible_dtypes and dtype2 in feasible_dtypes:
            return True
        
    return False


def preprocess_profiling_df(profiling_data: pd.DataFrame):

    profiling_data = preprocess_profiling_data(
        profiling_data=profiling_data,
        sample_limit=settings.STRATA_SAMPLE_LIMIT,
        dtypes_to_filter=["dimension"],
        truncate_sample_data=True
    )

    if settings.REMOVE_DATETIME_LP:
        profiling_data = profiling_data.loc[
            ~(profiling_data["datatype_l1"] == "date & time")
        ].reset_index(drop=True)

    condn2 = (
        profiling_data["datatype_l1"].isin(["integer", "float"])
    ) & (profiling_data["datatype_l2"] == "dimension")

    profiling_data["datatype"] = profiling_data["datatype_l1"]

    profiling_data.loc[condn2, "datatype"] = (
        profiling_data.loc[condn2, "datatype_l1"] + "_dimension"
    )

    condn = profiling_data["datatype_l1"] == "date & time"

    if condn.any():
        log.info(
            "[!] Warning date time should not be considered in link prediction"
        )
        profiling_data.loc[condn, "date_time_format"] = profiling_data.loc[
            condn, "sample_data"
        ].apply(
            classify_datetime_format,
        )

        def datetime_format(dtype, format):
            return f"{dtype} ({format})"

        profiling_data.loc[condn, "datatype"] = profiling_data.loc[condn].apply(
            lambda row: datetime_format(
                row["predicted_datatype_l1"], row["date_time_format"]
            ),
            axis=1,
        )
    
    def percent_conversion(x):
        return f"{(x * 100):.2f}%"

    profiling_data["uniqueness_ratio"] = profiling_data["uniqueness"]
    profiling_data["uniqueness"] = profiling_data.uniqueness.apply(
        percent_conversion
    )
    profiling_data["completeness"] = profiling_data.completeness.apply(
        percent_conversion
    )

    return profiling_data


import logging
import os

from intugle.core import settings

log = logging.getLogger(__name__)


def di_initalizer():
    MAX_RETRIES = 5
    par_vec_path = os.path.join(
    settings.MODEL_DIR_PATH, "dependant", "datatype_l1_identification"
)

    import nltk

    from .paragraph_vectors import (
        initialise_nltk,
    )

    nltk_local_path = os.path.join(par_vec_path, "nltk_data")
    nltk.data.path.append(nltk_local_path)

    count = MAX_RETRIES
    while count > 0:
        try:
            initialise_nltk()
            return
        except Exception as ex:
            log.error(
                f"[!] Error initating nltk , word embeddings and pretrained doc 2 vec model : {ex} ... \n\n Retries left {count}"
            )
            count -= 1


# ====================================================================
#  Importing the required python packages
# ====================================================================

import logging
import multiprocessing
import re

from collections import Counter, OrderedDict
from datetime import datetime

import numpy as np
import pandas as pd

from functional import pseq
from nltk.corpus import words
from trieregex import TrieRegEx as TRE

from intugle.core.settings import settings
from intugle.core.utilities.processing import compute_stats

from .initialize import di_initalizer

log = logging.getLogger(__name__)


try:
    x = words.words()
except Exception:
    di_initalizer()

words_list = []
for value in words.words():
    if len(value) > settings.DI_CONFIG["THRESHOLD"]["CORPUS_WORD_LIMIT"]:
        words_list.append(value.lower())

# ### TRIE REGEX for the words list
tre = TRE()
tre = TRE(*words_list)


# Get the core count and partition size for multiprocessing
core_count = multiprocessing.cpu_count()  # 1
size = settings.DI_CONFIG["THRESHOLD"]["PARTITION_SIZE"]

# Get the minimum and maximum range for the date extraction
minRange = settings.DI_CONFIG["THRESHOLD"]["DATE_MIN_YEAR"]
maxRange = settings.DI_CONFIG["THRESHOLD"]["DATE_MAX_YEAR"]


# ====================================================================
# checkInt:
#  - Check if the data is of Integer type or not
#  - If integer then return 1 else 0
# Parameters:
#     strs - input data (check if the data is integer or not)
# ====================================================================
def checkInt(strs) -> int:
    # If integer type then return 1 else return 0
    if isinstance(strs.replace(",", ""), int):
        return 1
    elif isinstance(
        strs.replace(",", ""), float
    ):  # If the data is matching for float return 0
        return 0
    else:
        try:
            int(strs.replace(",", ""))
            return 1
        except Exception:
            return 0


# ====================================================================
# checkFloat:
#  - Check if the data is of Float type or not
#  - If float then return 1 else 0
# Parameters:
#     strs - input data (check if the data is float or not)
# ====================================================================
def checkFloat(strs) -> int:
    # If Float type then return else return 0
    if isinstance(strs, float):
        return 1
    else:
        try:
            if checkInt(
                strs.replace(",", "")
            ):  # If the data is matching for Integer return 0
                return 0
            strs = float(strs.replace(",", ""))
            if strs != np.inf:  # If the data is matching for Infinity return 1
                return 1
            else:
                return 0
        except Exception:
            return 0


# ====================================================================
# alphaAndNumericMatch:
#  - Calculate the ratio of alpha to numeric ratio
#  - If the data is only alpha then returns alpha
#  - If the data is only numeric then return numeric
#  - If the data contains both alpha and numeric or special character then return alphanumeric
# Parameters:
#     value - input data
# ====================================================================


def otherSpecialCharacterCheck(value: str) -> int:
    """
    Handling special strings like 9.8.1 , 10.0.0 as alphanumperic
    Args:
        value (str): _description_

    Returns:
        int: _description_
    """
    # float pattern to also match edge floats like '9.' or '.5'
    float_pattern = re.compile(settings.DI_CONFIG["PREPROCESS_CONSTANT"]["FLOAT_PATTERN"])

    # Remove all float-like substrings
    text_without_floats = float_pattern.sub("", value)

    # Count remaining special characters
    special_chars = re.findall(
        pattern=settings.DI_CONFIG["PREPROCESS_CONSTANT"]["SPECIAL_CHARACTER_PATTERN_WITH_DOT"],
        string=text_without_floats,
    )
    return len(special_chars)


def alphaAndNumericMatch(value) -> str:
    # Converts the values into string type
    value = str(value)

    # Get the length of
    charCount = len(
        re.findall(
            string=value, pattern=settings.DI_CONFIG["PREPROCESS_CONSTANT"]["TEXT_PATTERN"]
        )
    )
    numCount = len(
        re.findall(
            string=value, pattern=settings.DI_CONFIG["PREPROCESS_CONSTANT"]["NUMBER_PATTERN"]
        )
    )
    specialCharCount = len(
        re.findall(
            string=value,
            pattern=settings.DI_CONFIG["PREPROCESS_CONSTANT"]["SPECIAL_CHAR_PATTERN"],
        )
    )
    otherspecialCharCount = otherSpecialCharacterCheck(value)

    # Based on the occurence of the characters and number return the respective types
    if (charCount > 0 or specialCharCount or otherspecialCharCount) and numCount > 0:
        return "alphanumeric"
    elif numCount > 0:
        return "numeric"
    elif charCount > 0:
        return "alpha"
    else:
        return "others"


# ====================================================================
# intTypeData:
#  - Iterate through each element and check if they are integer type
#  - Get the total length of characters in the integer data
#  - If the length of int type is 0, then return default values else compute the statistics
# Parameters:
#     col_values - list of input data
# ====================================================================
def intTypeData(col_values):
    # Iterate through each element and check if they are integer type
    int_type_data = [int(element) for element in col_values if checkInt(element)]
    int_type_ratio = len(int_type_data) / len(col_values)
    _median = np.median(int_type_data)

    # Get the total length of characters in the integer data
    try:
        mean_before_int = np.mean([len(str(val)) for val in int_type_data])
    except Exception:
        mean_before_int = 0

    # If the length of int type is 0, then return default values else compute the statistics
    if len(int_type_data) > 0:
        _mean, _variance, _skew, _kurtosis, _min, _max, _sum = compute_stats(
            int_type_data
        )
        return (
            int_type_data,
            int_type_ratio,
            _mean,
            _median,
            _variance,
            _skew,
            _kurtosis,
            _min,
            _max,
            _sum,
            mean_before_int,
        )
    else:
        return int_type_data, int_type_ratio, 0, 0, 0, 0, -3, 0, 0, 0, mean_before_int


# ====================================================================
# floatTypeData:
#  - Iterate through each element and check if they are float type
#  - Get the total length of characters, before and after decimal point in the float data
#  - Check if the max value is 0, then set the flag = 1 else 0
#  - If the length of float type is 0, then return default values else compute the statistics
# Parameters:
#     col_values - list of input data
# ====================================================================
def floatTypeData(col_values):
    # Iterate through each element and check if they are float type
    float_type_data = [float(element) for element in col_values if checkFloat(element)]
    float_type_ratio = len(float_type_data) / len(col_values)
    _median = np.median(float_type_data)

    # Get the total length of characters, before and after decimal point in the float data
    try:
        mean_before_float = np.mean(
            [
                len(str(float(val)).split(".")[0])
                for val in float_type_data
                if pd.notnull(val)
            ]
        )
        mean_after_float = np.mean(
            [
                len(str(float(val)).split(".")[1])
                for val in float_type_data
                if len(str(float(val)).split(".")) > 1 and pd.notnull(val)
            ]
        )
        max_after_float = max(
            [
                int(str(float(val)).split(".")[1])
                for val in float_type_data
                if len(str(float(val)).split(".")) > 1 and pd.notnull(val)
            ]
        )
    except Exception:
        mean_before_float = 0
        mean_after_float = 0
        max_after_float = 0

    # Check if the max value is 0, then set the flag = 1 else 0
    zero_flag = 1

    if max_after_float > 0:
        zero_flag = 0

    # If the length of float type is 0, then return default values else compute the statistics
    if len(float_type_data) > 0:
        _mean, _variance, _skew, _kurtosis, _min, _max, _sum = compute_stats(
            float_type_data
        )
        return (
            float_type_data,
            float_type_ratio,
            _mean,
            _median,
            _variance,
            _skew,
            _kurtosis,
            _min,
            _max,
            _sum,
            mean_before_float,
            mean_after_float,
            max_after_float,
            zero_flag,
        )
    else:
        return (
            float_type_data,
            float_type_ratio,
            0,
            0,
            0,
            0,
            -3,
            0,
            0,
            0,
            mean_before_float,
            mean_after_float,
            max_after_float,
            zero_flag,
        )


# ====================================================================
# checkDate:
#  - Validate if the input data is of date/datetime format
#  - Preprocess and check the string against the pattern and return flag
# Parameters:
#     strs - input data
# ====================================================================
def checkDate(strs) -> int:
    date_pattern = " \d{4}-[0-1][0-9]-[0-3][0-9](-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{4}-[0-3][0-9]-[0-1][0-9](-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | [0-1][0-9]-\d{4}-[0-3][0-9](-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | [0-3][0-9]-\d{4}-[0-1][0-9](-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | [0-1][0-9]-[0-3][0-9]-\d{4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | [0-3][0-9]-[0-1][0-9]-\d{4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-\d{1,2}-\d{4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{4}-\d{1,2}-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-\d{4}-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{2,4}-(\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{2,4}-\d{1,2}-(\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | (\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+-\d{2,4}-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)*  | (\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+-\d{1,2}-\d{2,4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)*  | \d{1,2}-(\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+-\d{2,4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-\d{2,4}-(\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)*  | \d{2,4}-(\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)*  | \d{2,4}-\d{1,2}-(\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | (\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+-\d{2,4}-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | (\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+-\d{1,2}-\d{2,4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-(\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+-\d{2,4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-\d{2,4}-(\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{2,4}-(\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{2,4}-(\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | (\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+-\d{2,4} (-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | (\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+-\d{2,4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-(\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-(\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | (\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | (\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+-\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{2,4}-(\\bjanuary\\b|\\bfebruary\\b|\\bmarch\\b|\\bapril\\b|\\bmay\\b|\\bjune\\b|\\bjuly\\b|\\baugust\\b|\\bseptember\\b|\\boctober\\b|\\bnovember\\b|\\bdecember\\b)+-\d{2,4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{2,4}-(\\bjan\\b|\\bfeb\\b|\\bmar\\b|\\bapr\\b|\\bmay\\b|\\bjun\\b|\\bjul\\b|\\baug\\b|\\bsep\\b|\\bsept\\b|\\boct\\b|\\bnov\\b|\\bdec\\b)+-\d{2,4}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}-days-\d{1,2}:\d{1,2}:\d{1,2}(-\d{1,2}:\d{1,2}(:\d{1,2})*(z)*[-\d{0,3}]*(AM|PM|CDT|EDT|IST)*)* | \d{1,2}h[-]*\d{1,2}m | \d{1,2}:\d{2}[:|-]+\d{2} | \d{1,2}:\d{2}[-]*(AM|PM|CDT|EDT|IST)* | (FY|FQ)+[-]*\d{2,4} | \d+[-]*(year[s]*|month[s]*|day[s]*|week[s]*|year[s]*|hour[s]|minute[s]|second[s])+ | \d{1,2}-\d{2}-\d{2}[-]*(AM|PM|CDT|EDT|IST)* | \d{1,2}:\d{2}-(AM|PM|CDT|EDT|IST)+-\d{1,2}:\d{2}-(AM|PM|CDT|EDT|IST)+ | \b(?:20\d{2}|19\d{2}|\d{2})[-/._](0[1-9]|1[0-2])[-/._](0[1-9]|[12]\d|3[01])T([01]\d|2[0-4])[:,.](0[0-9]|[1-5]\d)[:,.](0[0-9]|[1-5]\d)Z\b | \b(?:20\d{2}|19\d{2}|\d{2})[-/._](0[1-9]|[12]\d|3[01])[-/._](0[1-9]|1[0-2])T([01]\d|2[0-4])[:,.](0[0-9]|[1-5]\d)[:,.](0[0-9]|[1-5]\d)Z\b"

    # Preprocess the text for pattern matching
    strs = (
        str(strs)
        .replace("/", " ")
        .replace(",", " ")
        .replace(".", " ")
        .replace(" ", "-")
    )
    strs = re.sub(string=strs, pattern="-+", repl="-")

    # Match the text with date pattern and return 1 or 0 based on matching
    matched = re.match(string=" " + strs + " ", pattern=date_pattern, flags=re.I)
    return 1 if matched else 0


# ====================================================================
# checkOtherDate:
#  - Match the text with day pattern and return 1 or 0 based on matching
# Parameters:
#     strs - input data
# ====================================================================
def checkOtherDate(strs) -> int:
    # Pattern for the day checks
    days_abbr = [
        "saturday",
        "sunday",
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "october",
        "november",
        "december",
        "january",
        "february",
        "march",
        "april",
        "may",
        "june",
        "july",
        "august",
        "september",
    ]
    days_abbr_patt = "\\b" + "\\b|\\b".join(days_abbr) + "\\b"

    # Match the text with day pattern and return 1 or 0 based on matching
    day_check = re.findall(string=strs, pattern=days_abbr_patt, flags=re.I)
    return 1 if len(day_check) > 0 else 0


# ====================================================================
# checkDateRange:
#  - Match the date range and return 1 or 0 based on condition
# Parameters:
#     strs - input data
# ====================================================================
def checkDateRange(strs) -> int:
    try:
        if int(strs) >= minRange and int(strs) <= maxRange:
            return 1
        else:
            return 0
    except Exception:
        return 0


# ====================================================================
# checkRange:
#  - Match the range pattern and return the flag based on the match condition
# Parameters:
#     strs - input data
# ====================================================================
def checkRange(vals):
    range_data = re.findall(
        pattern=settings.DI_CONFIG["PREPROCESS_CONSTANT"]["RANGE_PATTERN"], string=str(vals)
    )
    if len(range_data) > 0 and float(range_data[0][1]) >= float(range_data[0][0]):
        return 1


# ====================================================================
# upper_char_len_in_cells:
#  -  Check the total number of upper case characters
# Parameters:
#     values - input data
# ====================================================================
def upper_char_len_in_cells(values: str) -> int:
    return len(re.findall("[A-Z]", values)) / len(values)


# ====================================================================
# lower_char_len_in_cells:
#  -  Check the total number of lower case characters
# Parameters:
#     values - input data
# ====================================================================
def lower_char_len_in_cells(values: str) -> int:
    return len(re.findall("[a-z]", values)) / len(values)


# ====================================================================
# get_word_length:
#  -  Check the word length
# Parameters:
#     val - input data
# ====================================================================
def get_word_length(val: str) -> int:
    return len(val.split(" "))


def get_char_length(val: str) -> int:
    """
    - Get character length of a value
    Parameters
    ----------
    val (str): column values as string
    Returns
    -------
    int: total character length

    """
    return len(str(val))


# ====================================================================
# lexical_matching:
#  - Check if the cleaned data is empty
#  - Code should be calculated at Element level and not at word level
#  - Match each text with the TRIE regex on nltk word list
#  - Get the total matched text with english dictionary to total data
# Parameters:
#     cleaned_data - list of input data
# ====================================================================
def lexical_matching(cleaned_data: list) -> float:
    final_score = []

    # Check if the cleaned data is empty
    if len(cleaned_data) == 0:
        return 0
    else:
        words_freq = Counter(cleaned_data)

        # Code should be calculated at Element level and not at word level
        for search_word in words_freq.keys():
            bool_score = 0
            for words in search_word.lower().split():
                if tre.has(words):
                    bool_score = 1
                    break
            final_score.append(bool_score * words_freq[search_word])

        # Get the total matched text with english dictionary to total data
        return sum(final_score) / len(cleaned_data)


# ====================================================================
# alphanum_flag_creation:
#  - Change in the feature for alphanumeric features
#  -
# Parameters:
#     values - list of input data
# ====================================================================
def alphanum_flag_creation(values, alpha, numeric):
    # Change in the feature for alphanumeric features
    if alpha == 1:
        return 1
    elif numeric == 1:
        return 0
    else:
        return lexical_matching(values)


# ====================================================================
# url_identification_flag:
#  - Check the length of total urls present in text
# Parameters:
#     text - input data
# ====================================================================
def url_identification_flag(text: str) -> list:
    return [len(re.findall(settings.DI_CONFIG["PREPROCESS_CONSTANT"]["URL_PATTERN"], str(text)))]


# ====================================================================
# additional_features:
#  - Subset values for the alphanumeric, alpha, numeric and others type
#  - Calculate the upper and lower case characters and ratio and mean values
#  - Iterate through each elements and get the int, float, range type, date data for feature creation
#  - Get the ratio of the various features based on total values
#  - Get the statistical values for the Integer, Float, Date type, word length
#  - Get the url length and alphanum dictionary feature for each values
# Parameters:
#   col_values - list of input data
#   date_samples - number of rows for the date patten matching
# ====================================================================
def additional_features(col_values: list, date_samples: int = 1000) -> list:
    # Define the values for storing the features
    numeric_type = []
    int_type = []
    float_type = []
    alpha_type = []
    alphanum_type = []
    others_type = []
    # upper_case = []
    # lower_case = []
    range_type = []
    date_type = []

    # Defining the default values for the feature creation
    (
        mean_before_int,
        mean_before_float,
        mean_after_float,
        max_after_float,
        zero_flag,
        mean_uppercase,
        mean_lowercase,
    ) = 0, 0, 0, 0, 0, 0, 0
    (
        int_mean,
        int_variance,
        int_skew,
        int_kurtosis,
        int_min,
        int_max,
        int_sum,
        int_median,
    ) = 0, 0, 0, -3, 0, 0, 0, 0
    (
        float_mean,
        float_variance,
        float_skew,
        float_kurtosis,
        float_min,
        float_max,
        float_sum,
        float_median,
    ) = 0, 0, 0, -3, 0, 0, 0, 0
    (
        alphaNumRatio,
        numericRatio,
        alphaRatio,
        otherRatio,
        dateRatio,
        intRatio,
        floatRatio,
    ) = 0, 0, 0, 0, 0, 0, 0
    (
        wordlen_mean,
        wordlen_variance,
        wordlen_skew,
        wordlen_kurtosis,
        wordlen_min,
        wordlen_max,
        wordlen_sum,
        wordlen_median,
    ) = 0, 0, 0, 0, 0, 0, 0, 0
    url_mean, alphanum_dict_ratio = 0, 0

    # subset the values which is not empty for the feature creation
    col_values = [values for values in col_values if len(values) > 0]
    total_vals = len(col_values)

    # Subset values for the alphanumeric, alpha, numeric and others type
    log.info("Custom feature creation alphaAndNumericMatch started:%s", datetime.now())
    alphaNum = pseq(
        map(alphaAndNumericMatch, col_values), processes=core_count, partition_size=size
    )
    alphaNum = list(alphaNum)
    alpha_type = list(filter(lambda item: item == "alpha", alphaNum))
    alphanum_type = list(filter(lambda item: item == "alphanumeric", alphaNum))
    numeric_type = list(filter(lambda item: item == "numeric", alphaNum))
    others_type = list(filter(lambda item: item == "others", alphaNum))

    # Calculate the upper case characters and ratio and mean values
    log.info(
        "Custom feature creation upper and lower characters started:%s", datetime.now()
    )
    upper_case_ratio = pseq(
        map(upper_char_len_in_cells, col_values),
        processes=core_count,
        partition_size=size,
    )
    upper_case_ratio = list(upper_case_ratio)
    mean_uppercase = np.mean(upper_case_ratio) if len(upper_case_ratio) > 0 else 0

    # Calculate the lower case characters and ratio and mean values
    lower_case_ratio = pseq(
        map(lower_char_len_in_cells, col_values),
        processes=core_count,
        partition_size=size,
    )
    lower_case_ratio = list(lower_case_ratio)
    mean_lowercase = np.mean(lower_case_ratio) if len(lower_case_ratio) > 0 else 0

    # Iterate through each elements and get the integer type data for feature creation
    log.info("Custom feature creation checkInt started:%s", datetime.now())
    int_data = pseq(
        map(checkInt, col_values), processes=core_count, partition_size=size
    )
    # int_type = [int(col_values[idx].replace(',', '')) for idx, val in enumerate(int_data) if val == 1]
    int_type = [
        int(col_values[idx].replace(",", ""))
        for idx, val in enumerate(int_data)
        if val == 1
    ]
    # int_type = list(int_type)
    int_type = np.array(int_type, dtype=np.float128)
    # Iterate through each elements and get the float type data for feature creation
    log.info("Custom feature creation checkFloat started:%s", datetime.now())
    float_data = pseq(
        map(checkFloat, col_values), processes=core_count, partition_size=size
    )
    float_type = [
        float(col_values[idx].replace(",", ""))
        for idx, val in enumerate(float_data)
        if val == 1
    ]
    # float_type = list(float_type)
    float_type = np.array(float_type, dtype=np.float128)

    # Iterate through each elements and get the range type data for feature creation
    log.info("Custom feature creation checkRange started:%s", datetime.now())
    range_data = pseq(
        map(checkRange, col_values), processes=core_count, partition_size=size
    )
    range_type = list(filter(lambda item: item == 1, range_data))
    range_type = list(range_type)

    # Iterate through each elements and get the date type data for feature creation
    log.info("Custom feature creation checkDate started:%s", datetime.now())
    sub_values = col_values[:date_samples]

    # Iterate through each elements and get the date type data for feature creation
    date_data = pseq(
        map(checkDate, sub_values), processes=core_count, partition_size=size
    )
    date_data = list(date_data)

    # Iterate through each elements and get the other date type data for feature creation
    log.info("Custom feature creation checkOtherDate started:%s", datetime.now())
    day_data = pseq(
        map(checkOtherDate, sub_values), processes=core_count, partition_size=size
    )
    day_data = list(day_data)

    # Iterate through each elements and get the date range data for feature creation
    log.info("Custom feature creation DateRange started:%s", datetime.now())
    daterange_data = pseq(
        map(checkDateRange, sub_values), processes=core_count, partition_size=size
    )
    daterange_data = list(daterange_data)

    date_type = [
        max(date_data[val], day_data[val], daterange_data[val])
        for val in range(len(date_data))
    ]

    # Get the ratio of the various features based on total values
    alphaNumRatio = len(alphanum_type) / total_vals if total_vals > 0 else 0
    numericRatio = len(numeric_type) / total_vals if total_vals > 0 else 0
    alphaRatio = len(alpha_type) / total_vals if total_vals > 0 else 0
    otherRatio = len(others_type) / total_vals if total_vals > 0 else 0
    dateRatio = np.mean(date_type) if total_vals > 0 else 0
    intRatio = len(int_type) / total_vals if total_vals > 0 else 0
    floatRatio = len(float_type) / total_vals if total_vals > 0 else 0
    rangeRatio = len(range_type) / total_vals if total_vals > 0 else 0

    log.info("Custom feature creation Integer features:%s", datetime.now())

    # Get the statistical values for the Integer type
    int_median = np.median(int_type)
    int_median = int_median if pd.notnull(int_median) else 0

    try:
        mean_before_int = np.mean([len(str(val)) for val in int_type])
    except Exception:
        # mean_before_int = np.NaN
        mean_before_int = np.nan

    mean_before_int = mean_before_int if pd.notnull(mean_before_int) else 0

    if len(int_type) > 0:
        int_mean, int_variance, int_skew, int_kurtosis, int_min, int_max, int_sum = (
            compute_stats(int_type)
        )

    int_mean = int_mean if pd.notnull(int_mean) else 0
    int_variance = int_variance if pd.notnull(int_variance) else 0
    int_skew = int_skew if pd.notnull(int_skew) else 0
    int_kurtosis = int_kurtosis if pd.notnull(int_kurtosis) else -3
    int_min = int_min if pd.notnull(int_min) else 0
    int_max = int_max if pd.notnull(int_max) else 0
    int_sum = int_sum if pd.notnull(int_sum) else 0

    log.info("Custom feature creation Float features:%s", datetime.now())

    # Get the statistical values for the Float type
    float_median = np.median(float_type)
    float_median = float_median if pd.notnull(float_median) else 0

    try:
        float_elem = [
            str(float(val)).split(".") for val in float_type if pd.notnull(val)
        ]
        mean_before_float = np.mean([len(val[0]) for val in float_elem])
        mean_after_float = np.mean([len(val[1]) for val in float_elem if len(val) > 1])
        max_after_float = max([float(val[1]) for val in float_elem if len(val) > 1])
    except Exception:
        # max_after_float = np.NaN
        max_after_float = np.nan
        mean_before_float, mean_after_float = 0, 0

    # Get the Float related values before and after float types
    orig_max_after_float = max_after_float
    max_after_float = max_after_float if pd.notnull(max_after_float) else 0

    if pd.isnull(orig_max_after_float) or max_after_float > 0:
        zero_flag = 0
    else:
        zero_flag = 1

    if len(float_type) > 0:
        (
            float_mean,
            float_variance,
            float_skew,
            float_kurtosis,
            float_min,
            float_max,
            float_sum,
        ) = compute_stats(float_type)

    float_mean = float_mean if pd.notnull(float_mean) else 0
    float_variance = float_variance if pd.notnull(float_variance) else 0
    float_skew = float_skew if pd.notnull(float_skew) else 0
    float_kurtosis = float_kurtosis if pd.notnull(float_kurtosis) else -3
    float_min = float_min if pd.notnull(float_min) else 0
    float_max = float_max if pd.notnull(float_max) else 0
    float_sum = float_sum if pd.notnull(float_sum) else 0

    # Get the statistical values for the word length
    log.info("Custom feature creation word length features:%s", datetime.now())
    word_len_data = pseq(
        map(get_word_length, col_values), processes=core_count, partition_size=size
    )
    word_len_data = list(word_len_data)

    if len(word_len_data) > 0:
        (
            wordlen_mean,
            wordlen_variance,
            wordlen_skew,
            wordlen_kurtosis,
            wordlen_min,
            wordlen_max,
            wordlen_sum,
        ) = compute_stats(word_len_data)
        wordlen_median = np.median(word_len_data)
        wordlen_max = np.max(word_len_data)

    # Get the statistical values for the character length
    char_len_data = pseq(
        map(get_char_length, col_values), processes=core_count, partition_size=size
    )
    char_len_data = list(char_len_data)
    if len(char_len_data) > 0:
        charlen_max = np.max(char_len_data)
    else:
        charlen_max = 0

    # Get the url length for each values
    log.info("Custom feature creation URL Identification features:%s", datetime.now())
    url_len_data = pseq(
        map(url_identification_flag, col_values),
        processes=core_count,
        partition_size=size,
    )
    url_len_data = list(url_len_data)
    url_mean = np.mean(url_len_data) if total_vals > 0 else 0

    # Alphanum Dictionary Flag feature Creation
    log.info(
        "Custom feature creation Alphanum Dictionary Flag features:%s", datetime.now()
    )
    alphanum_dict_ratio = alphanum_flag_creation(col_values, alphaRatio, numericRatio)

    log.info("Custom feature creation case based features:%s", datetime.now())

    return [
        total_vals,
        alphaNumRatio,
        numericRatio,
        alphaRatio,
        otherRatio,
        dateRatio,
        intRatio,
        floatRatio,
        rangeRatio,
        int_mean,
        int_variance,
        int_skew,
        int_kurtosis,
        int_min,
        int_max,
        int_sum,
        int_median,
        mean_before_int,
        float_mean,
        float_variance,
        float_skew,
        float_kurtosis,
        float_min,
        float_max,
        float_sum,
        float_median,
        zero_flag,
        mean_before_float,
        mean_after_float,
        max_after_float,
        mean_uppercase,
        mean_lowercase,
        wordlen_mean,
        wordlen_variance,
        wordlen_skew,
        wordlen_kurtosis,
        wordlen_min,
        wordlen_max,
        wordlen_sum,
        wordlen_median,
        url_mean,
        alphanum_dict_ratio,
        charlen_max,
    ]


# ====================================================================
# extract_addl_feats:
#  - Creating of additional/custom features
#  - Call the additional features for the input values
#  - Iterate through created features and store it in the feature dictionary
# Parameters:
#   col_values - list of input data
#   features - dictionary of features
# ====================================================================
def extract_addl_feats(col_values: list, features: OrderedDict):
    # Creating of additional/custom features
    feats_name = [
        "total_vals",
        "alphaNumRatio",
        "numericRatio",
        "alphaRatio",
        "otherRatio",
        "dateRatio",
        "intRatio",
        "floatRatio",
        "rangeRatio",
        "int_mean",
        "int_variance",
        "int_skew",
        "int_kurtosis",
        "int_min",
        "int_max",
        "int_sum",
        "int_median",
        "mean_before_int",
        "float_mean",
        "float_variance",
        "float_skew",
        "float_kurtosis",
        "float_min",
        "float_max",
        "float_sum",
        "float_median",
        "zero_flag",
        "mean_before_float",
        "mean_after_float",
        "max_after_float",
        "mean_uppercase",
        "mean_lowercase",
        "wordlen_mean",
        "wordlen_variance",
        "wordlen_skew",
        "wordlen_kurtosis",
        "wordlen_min",
        "wordlen_max",
        "wordlen_sum",
        "wordlen_median",
        "url_mean",
        "alphanum_dict_ratio",
        "charlen_max",
    ]

    # Call the additional features for the input values
    start_time = datetime.now()
    log.info("Custom feature creation started:%s", start_time)
    feats_list = additional_features(col_values)
    end_time = datetime.now()
    log.info("Custom feature creation completed:%s", end_time)
    log.info("Total time taken:%s", end_time - start_time)

    # Iterate through created features and store it in the feature dictionary
    for iters, name in enumerate(feats_name):
        features[name] = feats_list[iters]



import json
import os
import uuid

from typing import Any, Dict, Optional, Self

import pandas as pd
import yaml

from intugle.adapters.factory import AdapterFactory
from intugle.adapters.models import (
    BusinessGlossaryOutput,
    ColumnGlossary,
    ColumnProfile,
    DataTypeIdentificationL1Output,
    DataTypeIdentificationL2Input,
    DataTypeIdentificationL2Output,
    KeyIdentificationOutput,
    ProfilingOutput,
)
from intugle.common.exception import errors
from intugle.core import settings
from intugle.core.pipeline.business_glossary.bg import BusinessGlossary
from intugle.core.pipeline.datatype_identification.l2_model import L2Model
from intugle.core.pipeline.datatype_identification.pipeline import DataTypeIdentificationPipeline
from intugle.core.pipeline.key_identification.ki import KeyIdentificationLLM
from intugle.models.resources.model import Column, ColumnProfilingMetrics
from intugle.models.resources.source import Source, SourceTables

# FIXME add type
DataSetData = pd.DataFrame


class DataSet:
    """
    A container for the dataframe and all its analysis results.
    This object is passed from one pipeline step to the next.
    """

    def __init__(self, data: DataSetData, name: str):
        # The original, raw dataframe object (e.g., a pandas DataFrame)
        self.id = uuid.uuid4()
        self.name = name
        self.data = data

        # The factory creates the correct wrapper for consistent API access
        self.adapter = AdapterFactory().create(data)

        # A dictionary to store the results of each analysis step
        self.results: Dict[str, Any] = {}

    def profile_table(self) -> Self:
        """
        Profiles the table and stores the result in the 'results' dictionary.
        """
        self.results["table_profile"] = self.adapter.profile(self.data)
        return self

    def profile_columns(self) -> Self:
        """
        Profiles each column in the dataset and stores the results in the 'results' dictionary.
        This method relies on the 'table_profile' result to get the list of columns.
        """
        if "table_profile" not in self.results:
            raise RuntimeError("TableProfiler must be run before profiling columns.")

        table_profile: ProfilingOutput = ProfilingOutput.model_validate(self.results["table_profile"])
        self.results["column_profiles"] = {
            col_name: self.adapter.column_profile(
                self.data, self.name, col_name, table_profile.count, settings.UPSTREAM_SAMPLE_LIMIT
            )
            for col_name in table_profile.columns
        }
        return self

    def identify_datatypes_l1(self) -> "DataSet":
        """
        Identifies the data types at Level 1 for each column based on the column profiles.
        This method relies on the 'column_profiles' result.
        """
        if "column_profiles" not in self.results:
            raise RuntimeError("TableProfiler and ColumnProfiler must be run before data type identification.")

        column_profiles: dict[str, ColumnProfile] = self.results["column_profiles"]
        records = []
        for column_name, stats in column_profiles.items():
            records.append({"table_name": self.name, "column_name": column_name, "values": stats.dtype_sample})

        l1_df = pd.DataFrame(records)
        di_pipeline = DataTypeIdentificationPipeline()
        l1_result = di_pipeline(sample_values_df=l1_df)

        column_datatypes_l1 = [DataTypeIdentificationL1Output(**row) for row in l1_result.to_dict(orient="records")]

        for column in column_datatypes_l1:
            column_profiles[column.column_name].datatype_l1 = column.datatype_l1

        self.results["column_datatypes_l1"] = column_datatypes_l1
        return self

    def identify_datatypes_l2(self) -> "DataSet":
        """
        Identifies the data types at Level 2 for each column based on the column profiles.
        This method relies on the 'column_profiles' result.
        """
        if "column_profiles" not in self.results:
            raise RuntimeError("TableProfiler and ColumnProfiler must be run before data type identification.")

        column_profiles: dict[str, ColumnProfile] = self.results["column_profiles"]
        columns_with_samples = [DataTypeIdentificationL2Input(**col.model_dump()) for col in column_profiles.values()]

        column_values_df = pd.DataFrame([item.model_dump() for item in columns_with_samples])
        l2_model = L2Model()
        l2_result = l2_model(l1_pred=column_values_df)
        column_datatypes_l2 = [DataTypeIdentificationL2Output(**row) for row in l2_result.to_dict(orient="records")]

        for column in column_datatypes_l2:
            column_profiles[column.column_name].datatype_l2 = column.datatype_l2

        self.results["column_datatypes_l2"] = column_datatypes_l2
        return self

    def identify_keys(self) -> Self:
        """
        Identifies potential primary keys in the dataset based on column profiles.
        This method relies on the 'column_profiles' result.
        """
        if "column_datatypes_l1" not in self.results or "column_datatypes_l2" not in self.results:
            raise RuntimeError("DataTypeIdentifierL1 and L2 must be run before KeyIdentifier.")

        column_profiles: dict[str, ColumnProfile] = self.results["column_profiles"]
        column_profiles_df = pd.DataFrame([col.model_dump() for col in column_profiles.values()])

        ki_model = KeyIdentificationLLM(profiling_data=column_profiles_df)
        ki_result = ki_model()
        output = KeyIdentificationOutput(**ki_result)
        key = output.column_name

        if key is not None:
            self.results["key"] = key
        return self

    def profile(self) -> Self:
        """
        Profiles the dataset including table and columns and stores the result in the 'results' dictionary.
        This is a convenience method to run profiling on the raw dataframe.
        """
        if self.data.empty:
            raise ValueError("The raw dataframe is empty. Cannot perform profiling.")
        self.profile_table().profile_columns()
        return self

    def identify_datatypes(self) -> Self:
        """
        Identifies the data types for the dataset and stores the result in the 'results' dictionary.
        This is a convenience method to run data type identification on the raw dataframe.
        """
        if self.data.empty:
            raise ValueError("The raw dataframe is empty. Cannot perform data type identification.")
        self.identify_datatypes_l1().identify_datatypes_l2()
        return self

    def generate_glossary(self, domain: str = "") -> Self:
        """
        Generates a business glossary for the dataset and stores the result in the 'results' dictionary.
        This method relies on the 'column_datatypes_l1' results.
        """
        if "column_datatypes_l1" not in self.results:
            raise RuntimeError("DataTypeIdentifierL1  must be run before Business Glossary Generation.")

        column_profiles: dict[str, ColumnProfile] = self.results["column_profiles"]
        column_profiles_df = pd.DataFrame([col.model_dump() for col in column_profiles.values()])

        bg_model = BusinessGlossary(profiling_data=column_profiles_df)
        table_glossary, glossary_df = bg_model(table_name=self.name, domain=domain)
        columns_glossary = []
        for _, row in glossary_df.iterrows():
            columns_glossary.append(
                ColumnGlossary(
                    column_name=row["column_name"],
                    business_glossary=row.get("business_glossary", ""),
                    business_tags=row.get("business_tags", []),
                )
            )
        glossary_output = BusinessGlossaryOutput(
            table_name=self.name, table_glossary=table_glossary, columns=columns_glossary
        )

        for column in glossary_output.columns:
            column_profiles[column.column_name].business_glossary = column.business_glossary
            column_profiles[column.column_name].business_tags = column.business_tags

        self.results["business_glossary_and_tags"] = glossary_output
        self.results["table_glossary"] = glossary_output.table_glossary
        return self


    def run(self, domain: str, save: bool = True) -> Self:
        """Run all stages """

        self.profile()\
            .identify_datatypes()\
            .identify_keys()\
            .generate_glossary(domain=domain)
        
        if save:
            self.save_yaml()

        return self

    # FIXME - this is a temporary solution to save the results of the analysis
    # need to use model while executing the pipeline
    def save_yaml(self, file_path: Optional[str] = None) -> None:
        if file_path is None:
            file_path = f"{self.name}.yml"
        file_path = os.path.join(settings.PROJECT_BASE, file_path)

        column_profiles = self.results.get("column_profiles")

        table_description = self.results.get("table_glossary")
        table_tags = self.results.get("business_glossary_and_tags")

        if column_profiles is None or table_description is None or table_tags is None:
            raise errors.NotFoundError(
                "Column profiles not found in the dataset results. Ensure profiling steps were executed."
            )

        columns: list[Column] = []

        for column_profile in column_profiles.values():
            column_profile = ColumnProfile.model_validate(column_profile)
            column = Column(
                name=column_profile.column_name,
                description=column_profile.business_glossary,
                type=column_profile.datatype_l1,
                category=column_profile.datatype_l2,
                tags=column_profile.business_tags,
                profiling_metrics=ColumnProfilingMetrics(
                    count=column_profile.count,
                    null_count=column_profile.null_count,
                    distinct_count=column_profile.distinct_count,
                    sample_data=column_profile.sample_data,
                ),
            )
            columns.append(column)

        table = SourceTables(name=self.name, description=table_description, columns=columns)

        source = Source(name="healthcare", description=table_description, schema="public", database="", table=table)

        sources = {"sources": [json.loads(source.model_dump_json())]}

        # Save the YAML representation of the sources
        with open(file_path, "w") as file:
            yaml.dump(sources, file, sort_keys=False, default_flow_style=False)

    def result_to_pandas(self):
        column_profiles = self.results.get("column_profiles")
        if column_profiles is None:
            return "<p>No column profiles available.</p>"
        df = pd.DataFrame([col.model_dump() for col in column_profiles.values()])
        return df

    def _repr_html_(self):
        df = self.result_to_pandas().head()
        return df._repr_html_()



import re

from typing import Callable, Optional

from pydantic import TypeAdapter

from intugle.libs.query_generator import QueryGenerator, QueryGeneratorModel

from .custom_data_types.OrderedSet import OrderedSet
from .models.models import (
    CategoryType,
    ETLModel,
    FieldDetailsModel,
    FieldsModel,
    FieldType,
    JoinOpt,
    LinkModel,
)
from .utils.cte import CTE
from .utils.filter import Filter
from .utils.groupby import GroupBy
from .utils.join import Join

# Regular expression to match the pattern @{number}[Ftext]
SQL_CODE_REGEX_PATTERN = r"@\{([\d\-]+)\}\[([\w\-\.\`]+)\]"
UNIQUENESS_THRESHOLD = 1
SEPERATOR = "____"


class SmartQueryGenerator:
    def __init__(
        self,
        etl: ETLModel,
        field_details_fetcher: Callable[[list[int]], dict[int, FieldDetailsModel]],
        links: list[LinkModel],
        field_details: dict[int, FieldDetailsModel],
        can_view_pii: bool = True,
        uniqueness_threshold: int = UNIQUENESS_THRESHOLD,
        sql_code_regex_pattern=SQL_CODE_REGEX_PATTERN,
    ):
        self.__etl = ETLModel.model_validate(etl)
        self.__field_details_fetcher = field_details_fetcher
        self.__can_view_pii = can_view_pii

        ta = TypeAdapter(list[LinkModel])
        self.__links = ta.validate_python(links)

        self.__field_details = field_details

        self.__ctes: list[tuple[Optional[str], str]] = []
        self.__assets: OrderedSet[int] = OrderedSet()
        self.__assets_join_opt = {}
        self.__join: dict[str, JoinOpt] = {}

        self.__selected_fields = []
        self.__join_json = {}
        self.__filters = {}
        self.__group_by = {}
        self.__sort_fields = []
        self.__limit = None
        self.__offset = None

        self.__data_categories = {"dimension": [], "measure": []}

        self.__sql_code_regex_pattern = sql_code_regex_pattern
        self.__uniqueness_threshold = uniqueness_threshold

    def __fetch_fields_details(self, fields: list[int]):
        __field_details = self.__field_details_fetcher(fields)

        self.__field_details = {**self.__field_details, **__field_details}

    def __get_join_names(self, join: dict):
        assets = {}
        for field in self.__field_details.values():
            if field.type == FieldType.source:
                assets[field.asset_id] = {
                    "id": field.asset_id,
                    "name": field.asset_name,
                }

        _tmp_join = join
        for _key, _value in join.items():
            _dataset = assets.get(_value["dataset_id"])
            if not _dataset:
                continue
            _tmp_join[_key]["dataset"] = _dataset["name"]
            fields = []
            if "fields" not in _value:
                continue
            for field in _value["fields"]:
                field = LinkModel.model_validate(field)
                # if not field.is_selected:
                #     continue

                left_field = self.__field_details.get(field.source_field_id)
                right_field = self.__field_details.get(field.target_field_id)

                fields.append({
                    "left_field_id": left_field.id,
                    "left_dataset_id": left_field.asset_id,
                    "left_field": left_field.sql_code,
                    "left_dataset": left_field.asset_name,
                    "left_sql_code": left_field.sql_code,
                    "right_field_id": right_field.id,
                    "right_dataset_id": right_field.asset_id,
                    "right_field": right_field.sql_code,
                    "right_dataset": right_field.asset_name,
                    "right_sql_code": right_field.sql_code,
                    **field.model_dump(),
                })
            _tmp_join[_key]["fields"] = fields

        return _tmp_join

    # GENERATE CALCULATED SQL CODES AND UPDATE SQL CODE IN FIELD ASSETS
    def __update_sql_code(self, field_id: int):
        field = self.__field_details[field_id]
        sql_code = field.sql_code

        # Find all matches
        matches = re.findall(self.__sql_code_regex_pattern, sql_code)

        if len(matches) <= 0:
            return sql_code

        # Use re.sub with a counter to replace each match with the corresponding element from the replacements list
        for match in matches:
            _field_id = int(match[0])
            replacement_sql_code = self.__update_sql_code(_field_id)
            sql_code = re.sub(
                self.__sql_code_regex_pattern,
                replacement_sql_code,
                sql_code,
                count=1,
            )

        self.__field_details[field_id].sql_code = sql_code
        return sql_code

    def __add_assets(self, field: FieldsModel, field_details: FieldDetailsModel, overwite: bool = False):
        asset_id = field_details.asset_id
        self.__assets.add(asset_id)
        if field.join_opt is not None and (overwite or asset_id not in self.__assets_join_opt):
            self.__assets_join_opt[str(asset_id)] = field.join_opt

    def __update_calcuated_field_base_field(self, field_id: int, base_field: FieldsModel):
        """
        Converts recursive calculated_fields to base fields
        Base field is used to pass the join options
        """
        field = self.__field_details[field_id]
        sql_code = field.sql_code

        if field.type != FieldType.calculated_field:
            self.__add_assets(base_field, field)

        matches = re.findall(self.__sql_code_regex_pattern, sql_code)

        # FIXME make it better @{}[]
        if len(matches) <= 0:
            return f"@{{{field.id}}}[{field.name}]"

        for match in matches:
            _field_id = int(match[0])
            # FIXME make it better @{}[]
            pattern = f"@{{{match[0]}}}[{match[1]}]"
            replacement_sql_code = self.__update_calcuated_field_base_field(_field_id, base_field)
            sql_code = sql_code.replace(pattern, replacement_sql_code)

        self.__field_details[field_id].sql_code = f"{sql_code}"

        return f"({sql_code})"

    def __add_field(self, field: FieldsModel):
        field_id = field.id
        field_detail = self.__field_details[field_id]

        sql_code = field.sql_code
        if sql_code is None:
            sql_code = field_detail.sql_code

        selected_field = {
            "dataset": field_detail.asset_name,
            "field": field_detail.name,
            "sql_code": sql_code,
            "measure_func": field.measure_func,
            "renameCol": field.name,
            "dimension_func": field.dimension_func,
        }

        if field_detail.is_pii and not self.__can_view_pii:
            selected_field["isFunction"] = True
            selected_field["function"] = [
                {
                    "funName": "custom",
                    "params": {"expr": f"MD5({field_detail.sql_code})"},
                }
            ]

        # FIXME Need to update query generator Needed function on function for this functionality
        if field.dimension_func is not None:
            selected_field["isFunction"] = True
            selected_field["function"] = [
                {
                    "funName": field.dimension_func,
                    "params": {"expr": field_detail.sql_code},
                }
            ]

        # Override if measure
        if field.category != CategoryType.measure:
            self.__data_categories["dimension"].append(field_id)
        else:
            self.__data_categories["measure"].append(field_id)
            selected_field["isFunction"] = True
            selected_field["function"] = [
                {
                    "funName": field.measure_func,
                    "params": {"expr": field_detail.sql_code},
                }
            ]

        self.__selected_fields.append(selected_field)

    def build_cte(self):
        cte_builder = CTE(
            self.__join,
            self.__etl,
            self.__field_details,
            SmartQueryGenerator,
            self.__field_details_fetcher,
            self.__links,
            self.__can_view_pii,
            self.__uniqueness_threshold,
            self.__sql_code_regex_pattern,
        )
        cte_builder.build()

        self.__ctes = cte_builder.get_ctes()
        self.__etl = cte_builder.get_etl()
        self.__field_details = cte_builder.get_field_details()
        self.__links = cte_builder.get_links()

    def build_assets_filter_sort(self):
        _filter = Filter(self.__etl.filter, self.__field_details, self.__can_view_pii)
        _filter.build_filters()
        assets = _filter.get_assets()
        for asset in assets:
            self.__assets.add(asset)

        if self.__etl.filter is not None:
            for sort in self.__etl.filter.sort_by:
                field = self.__field_details[sort.id]
                self.__assets.add(field.asset_id)

    def build_calculated_field_base_field(self):
        for field in self.__etl.fields:
            if field.type == FieldType.calculated_field:
                self.__update_calcuated_field_base_field(field.id, field)
            else:
                _f = self.__field_details[field.id]
                self.__add_assets(field, _f, True)

    def build_calculated_field_sql_codes(self):
        for field in self.__etl.fields:
            self.__update_sql_code(field.id)

    def build_fields(self):
        for field in self.__etl.fields:
            self.__add_field(field)

    def build_join(self, join: Optional[dict] = None):
        if join is None:
            for _field in self.__etl.fields:
                if _field.sql_code is not None:
                    continue
                _f = self.__field_details[_field.id]

                # if _f.type != FieldType.calculated_field:
                #     self.__add_assets(_field, _f, True)

            selected_fields = set(self.__field_details.keys())
            _join = Join(self.__links, selected_fields)
            join = _join.get_join_json(list(self.__assets), self.__assets_join_opt)

        fields = Join.get_fields(join, self.__links)

        for field in fields:
            field_details_keys = self.__field_details.keys()
            if field not in field_details_keys:
                self.__fetch_fields_details(fields)
                break

        self.__join = join

    def build_join_json(self):
        __join_json = self.__get_join_names(self.__join)
        self.__join_json = __join_json

    def build_filters(self):
        _filter = Filter(self.__etl.filter, self.__field_details, self.__can_view_pii)
        self.__filters = _filter.get_fresh_filter()

    def build_group_by(self):
        group_by = GroupBy(self.__etl, self.__field_details, self.__data_categories)
        group_by.build()
        self.__group_by = group_by.get_group_by()

    def build_sort_fields(self):
        sort_by = []

        # fields: dict[FieldsModel] = {}
        # for _field in self.__etl.fields:
        #     fields[_field.id] = _field

        for sort in self.__etl.filter.sort_by:
            sort_by_json = {
                "direction": sort.direction,
                # "renameCol": fields[sort.id].name,
                # "renameCol": sort.alias,
            }

            if sort.alias is not None:
                sort_by_json["renameCol"] = sort.alias
            else:
                sort_by_json["dataset"] = self.__field_details[sort.id].asset_name
                sort_by_json["field"] = self.__field_details[sort.id].name

            sort_by.append(sort_by_json)

        self.__sort_fields = sort_by

    def build_limit(self):
        self.__limit = self.__etl.filter.limit

    def build_offset(self):
        self.__offset = self.__etl.filter.offset

    def build(self, cte=True):
        if len(self.__etl.fields) <= 0:
            return
        self.build_calculated_field_base_field()
        self.build_assets_filter_sort()
        self.build_join(self.__etl.join)
        if cte:
            self.build_cte()
        self.build_calculated_field_sql_codes()
        self.build_join_json()
        self.build_fields()
        self.build_group_by()

        if self.__etl.filter is not None:
            self.build_filters()
            self.build_sort_fields()
            self.build_limit()
            self.build_offset()

    def __generate_query(self):
        query_generate_json = QueryGeneratorModel(
            selected_fields=self.__selected_fields,
            join=self.__join_json,
            filters=self.__filters,
            groupby=self.__group_by,
            sort_fields=self.__sort_fields,
            limit=self.__limit,
            offset=self.__offset,
        )

        query = QueryGenerator().getQuery(query_generate_json)
        return query

    def prepare_query(self):
        self.__ctes.append((None, self.__generate_query()))

    def get_query(self):
        self.prepare_query()

        if len(self.__ctes) == 1:
            _, _query = self.__ctes[0]
            return _query

        final_query = ""
        ctes = []
        caller_query = ""
        for _name, _query in self.__ctes:
            if _name is None:
                caller_query = _query
                continue
            ctes.append(f"{_name} as ( {_query} )")

        final_query = f"WITH {', '.join(ctes)} {caller_query}"

        return final_query

    def get_join(self):
        return self.__join

    def get_field_details(self):
        return self.__field_details



from typing import Optional

from pydantic import BaseModel, Field

from .operators import Operators
from .transformation import Transformations
from .utils import get_formatted_dataset_field


class QueryGeneratorModel(BaseModel):
    selected_fields: list[dict] = Field(default_factory=list)
    join: dict = Field(default_factory=dict)
    filters: dict = Field(default_factory=dict)
    groupby: dict = Field(default_factory=dict)
    sort_fields: list[dict] = Field(default_factory=list)
    limit: Optional[int] = None
    offset: Optional[int] = None


# @dataclass_json
# @dataclass
# class QueryGeneratorConfig:
#     selected_fields: list[dict] = field(default_factory=list)
#     join: dict = field(default_factory=dict)
#     filters: dict = field(default_factory=dict)
#     groupby: dict = field(default_factory=dict)
#     sort_fields: list[dict] = field(default_factory=list)
#     limit: Optional[int] = None
#     offset: Optional[int] = None


class QueryGenerator:
    @classmethod
    def get_select_fields(cls, selectFields):
        # Extracting Select fields
        selectedFields = []

        if len(selectFields) == 0:
            selectedFieldsExpr = " * "
        else:
            for field in selectFields:
                isFunction = field.get("isFunction", None)
                if isFunction:
                    #         func = functions[field["function"][0]["funName"]]
                    func = getattr(Transformations, field["function"][0]["funName"])
                    expr = func(**field["function"][0]["params"])
                else:
                    expr = field.get("sql_code", None)
                    if expr is None:
                        expr = get_formatted_dataset_field(
                            field["dataset"], field["field"]
                        )

                renameCol = field.get("renameCol", None)
                if renameCol:
                    expr = f"{expr} as `{renameCol}`"

                selectedFields.append(expr)
            selectedFieldsExpr = ", ".join(selectedFields)

        return selectedFieldsExpr

    @classmethod
    def get_join_expr(cls, joinFields):
        join_query = " FROM "
        # if len(joinFields) == 1:
        #     join_query += f" {joinFields['0']['dataset']} "
        #     return join_query
        tbl_cnt = len(joinFields)
        for _i in range(tbl_cnt):
            i = str(_i)
            if _i == 0:
                join_query += f"`{joinFields[i]['dataset']}`"
            else:
                joining_cndns = []
                for join_col in joinFields[i]["fields"]:
                    left_sql_code = join_col.get("left_sql_code", None)
                    right_sql_code = join_col.get("right_sql_code", None)
                    if left_sql_code is None:
                        left_sql_code = get_formatted_dataset_field(
                            join_col["left_dataset"], join_col["left_field"]
                        )
                    if right_sql_code is None:
                        right_sql_code = get_formatted_dataset_field(
                            join_col["right_dataset"], join_col["right_field"]
                        )
                    joining_cndns.append(left_sql_code + " = " + right_sql_code)

                joining_cndn = " AND ".join(joining_cndns)

                join_query += (
                    " "
                    + joinFields[i].get("join_type", "").upper()
                    + " JOIN "
                    + f"`{joinFields[i]['dataset']}`"
                    + " ON "
                    + str(joining_cndn)
                )
        return join_query

    def _filter(self, filterFields):
        fields = filterFields.get("fields", [])
        filterCols = []
        for _field in fields:
            _fields = _field.get("fields", None)
            if _fields:
                expr = self._filter(self, _field)
                expr = f"({expr})"
            else:
                expr = Operators.operator_factory(_field)
            filterCols.append(expr)
        return " {} ".format(filterFields["condition"].upper()).join(filterCols)

    @classmethod
    def get_filter_expr(cls, filterFields):
        # Filter Expression
        filterExpr = ""
        if filterFields != {}:
            filterCols = cls._filter(cls, filterFields)
            if filterCols:
                filterExpr = filterExpr + " WHERE " + filterCols
        return filterExpr

    @classmethod
    def get_groupby_expr(cls, groupbyFields):
        groupbyExpr = ""
        if groupbyFields != {} and len(groupbyFields["keys"]) != 0:
            groupbyCols = []
            for field in groupbyFields["keys"]:
                # _dataset = field.get("dataset", None)
                # _field = field["field"]
                # if _dataset is not None:
                #     _dataset_field = get_formatted_dataset_field(_dataset, _field)
                # else:
                #     _dataset_field = f"`{_field}`"

                # _func = field.get("function", None)
                # if _func is not None:
                #     # FIXME temp fix
                #     _dataset_field = f"{_func}({_dataset_field})"
                renameCol = field.get("renameCol", None)
                if renameCol is not None:
                    _dataset_field = f"`{renameCol}`"
                else:
                    _dataset_field = field["sql_code"]
                groupbyCols.append(_dataset_field)

            groupbyExpr = " GROUP BY " + " , ".join(groupbyCols)

            if "having" in groupbyFields and len(groupbyFields["having"]) != 0:
                groupbyExpr += cls.get_filter_expr(groupbyFields["having"]).replace(
                    " WHERE ", " HAVING ", 1
                )
        return groupbyExpr

    @classmethod
    def get_sort_expr(cls, sortFields):
        # Sort Expression
        sortExpr = ""
        if sortFields != []:
            sortExpr = " ORDER BY "
            sortCols = []
            for sortField in sortFields:
                renameCol = sortField.get("renameCol", None)
                if renameCol is not None:
                    sortCol = f"`{renameCol}`"
                else:
                    sortCol = get_formatted_dataset_field(
                        sortField["dataset"], sortField["field"]
                    )

                direction = sortField.get("direction", "asc")
                if direction == "asc" or direction == "ascending":
                    sortOp = "ASC"
                elif direction == "desc" or direction == "descending":
                    sortOp = "DESC"
                else:
                    sortOp = ""
                sortCols.append(sortCol + " " + sortOp)
            sortExpr = sortExpr + " , ".join(sortCols)
        return sortExpr

    @classmethod
    def get_limit_expr(cls, limit):
        return " LIMIT {}".format(limit)

    @classmethod
    def get_offset_expr(cls, offset):
        return " OFFSET {}".format(offset)

    @classmethod
    def getQuery(cls, config: QueryGeneratorModel):
        sqlQuery = ""

        if config.selected_fields != []:
            sqlQuery += "SELECT "
            sqlQuery += cls.get_select_fields(config.selected_fields)
        else:
            return sqlQuery

        # From dataset
        # if "from" in config:
        #     sqlQuery = sqlQuery + " FROM " + f"`{config['from']}`"

        if config.join != {}:
            sqlQuery += cls.get_join_expr(config.join)

        if config.filters != {}:
            sqlQuery += cls.get_filter_expr(config.filters)

        if config.groupby != {}:
            sqlQuery += cls.get_groupby_expr(config.groupby)

        if config.sort_fields != []:
            sqlQuery += cls.get_sort_expr(config.sort_fields)

        if config.limit is not None:
            sqlQuery += cls.get_limit_expr(config.limit)

        if config.offset is not None:
            sqlQuery += cls.get_offset_expr(config.offset)

        return sqlQuery



from .transformation import Transformations
from .utils import get_formatted_dataset_field


class Operators:
    @classmethod
    def _in(cls, field, value, **_):
        return "{} IN {}".format(field, value)

    @classmethod
    def _not_in(cls, field, value, **_):
        return "{} NOT IN {}".format(field, value)

    @classmethod
    def _like(cls, field, value, **_):
        return "{} LIKE {}".format(field, value)

    @classmethod
    def _equals(cls, field, value, **_):
        return "{} = {}".format(field, value)

    @classmethod
    def _not_equals(cls, field, value, **_):
        return "{} != {}".format(field, value)

    @classmethod
    def _greater_than(cls, field, value, **_):
        return "{} > {}".format(field, value)

    @classmethod
    def _less_than(cls, field, value, **_):
        return "{} < {}".format(field, value)

    @classmethod
    def _greater_than_equal_to(cls, field, value, **_):
        return "{} >= {}".format(field, value)

    @classmethod
    def _less_than_equal_to(cls, field, value, **_):
        return "{} <= {}".format(field, value)

    @classmethod
    def _is_null(cls, field, **_):
        return "{} IS NULL".format(field)

    @classmethod
    def _is_not_null(cls, field, **_):
        return "{} IS NOT NULL".format(field)

    @classmethod
    def _between(cls, field, lowerValue, upperValue, **_):
        return "{} BETWEEN {} AND {}".format(field, lowerValue, upperValue)

    # FIXME is too messy review logic
    @classmethod
    def operator_factory(cls, filterField):
        md5_hash = filterField.get("md5_hash", False)
        params = filterField["params"]
        func = getattr(cls, "_" + filterField["operator"])
        typ = params.get("type", "string")
        isFunction = filterField.get("isFunction", False)
        function = filterField.get("function", None)

        if typ == "dataset_field":
            params["value"] = params["value"].get("sql_code", None)
            if params["value"] is None:
                params["value"] = get_formatted_dataset_field(
                    params["value"]["dataset"], params["value"]["dataset"]
                )
        if typ == "string":
            value = params.get("value", "")
            params["value"] = f"'{value}'"

        dataset_field = filterField.get("sql_code", None)
        if dataset_field is None:
            dataset_field = get_formatted_dataset_field(
                filterField["dataset"], filterField["field"]
            )

        # if fieldType == "calculated_field":
        #     dataset_field = filterField["field"]
        # else:
        #     dataset_field = get_formatted_dataset_field(filterField["dataset"], filterField["field"])

        if isFunction is True and function is not None:
            # func = functions[field["function"][0]["funName"]]
            transFunc = getattr(Transformations, function[0]["funName"])
            dataset_field = transFunc(**function[0]["params"])

        if md5_hash:
            dataset_field = f"MD5({dataset_field})"

        return func(dataset_field, **params)



# ====================================================================
#  Importing the required python packages
# ====================================================================

# from core.conf import
import logging
import re

import pandas as pd

from langchain.output_parsers import ResponseSchema
from tqdm.auto import tqdm

from intugle.core.llms.chat import ChatModelLLM
from intugle.core.settings import settings
from intugle.core.utilities.processing import adjust_sample

log = logging.getLogger(__name__)

tqdm.pandas()


LLM_SAMPLE_SIZE = settings.DI_CONFIG["L2PARAMS"]["LLM_SAMPLE_SIZE"]


class L2ModelLLM:
    """
    L2 model is mainly responsible for classifying:
    - the integer and float datatype into integer_measure or
    - integer_dimension and same for float.
    - this model also finds the specific format the date & time datatype lies in (e.g. YYYY:MM:DD , YYYY-MM-DD, YYYY:MM:DD H:m:s , etc...)
    - this model uses llm for dimension & measure classification and a regex based approach for the date & time classification.
    """

    DIM_MEASURE_PROMPT = """Class Category
    - Dimensions contain qualitative information. These are descriptive attributes, like a product category, product key, customer address, or country. Dimensions can contain numeric characters (like an alphanumeric customer ID), but are not numeric values (It wouldn’t make sense to add up all the ID numbers in a column, for example).
    - Measures contain quantitative values that you can measure. Measures can be aggregated.
    Please analyze the table and classify column as either a 'Measure' or a 'Dimension' or 'Unknown'.
    INPUT
    ### TABLE:\n{table}\n\n 
    ### COLUMN NAME: {column_name}\n
    ### CLASS CATEGORY:{format_instructions}
    """
    dm_class_schema = [
        ResponseSchema(
            name="DM_class",
            description="This final class of a column Dimension, Measure & Unknown",
        )
    ]

    LLM_CONFIG = {"temperature": 0.0}

    def __init__(self, *args, **kwargs):
        # langfuse for monitoring

        self.chat_llm = ChatModelLLM.build(
            model_name=settings.LLM_PROVIDER,
            llm_config=self.LLM_CONFIG,
            template_string=self.DIM_MEASURE_PROMPT,
            response_schemas=self.dm_class_schema,
        )

        self.table_dict = {}

    def __classify_dim_measure__(self, table: str, column_name: str) -> str:
        response, parsing_success, _ = self.chat_llm.invoke(
            table=table,
            column_name=column_name,
            metadata={"column": f"{table}.{column_name}"},
        )

        if parsing_success:
            return response["DM_class"]
        else:
            try:
                pattern = re.compile(r'"DM_class":\s+"(Dimension|Measure|Unknown)"')
                return re.findall(pattern, response)[-1]
            except Exception as ex:
                log.error(f"[!] Error while parsing: {ex}")
                return "Unknown"

    def __call__(self, row) -> str:
        column_name = row["column_name"]

        sample_data = adjust_sample(
            sample_data=row["sample_data"], expected_size=settings.L2_SAMPLE_LIMIT
        )

        table = pd.DataFrame(sample_data, columns=[column_name])

        value = self.__classify_dim_measure__(table=str(table), column_name=column_name)

        return value.lower()


class L2Model:
    def __init__(self, *args, **kwargs):
        self.__model = L2ModelLLM(*args, **kwargs)

    def __call__(
        self,
        l1_pred: pd.DataFrame,
    ):

        l1_pred["predicted_datatype_l2"] = l1_pred.progress_apply(
            self.__model,
            axis=1,
        )

        return l1_pred



import textwrap

from intugle.mcp.semantic_layer.schema import SQLDialect


class Prompts:
    """
    LLM Prompts used by executor server
    """

    @classmethod
    def raw_executor_prompt(
        cls,
        dialect: SQLDialect,
        domain: str = "Industry",
        universal_instructions: str = "",
    ) -> str:
        """
        Given a dialect returns the appropriate executor agent prompt ( prompt for handling)
        Args:
            dialect (SQLDialect): _description_

        Returns:
            str: _description_
        """

        prompt = textwrap.dedent("""
        You are a Business Intelligence (BI) assistant with access to a {source} database from {domain}. Your role is to help users explore the data, extract useful insights, and answer their questions using available tools.

        Tools:
        - `get_tables`: Retrieve a list of all available tables.
        - `get_schema`: Get the schema details for a specified tables
        - `execute_query`: Run queries using the specified {dialect} dialect.
        - `get_user_guidance`: Access relevant examples and helpful logic for building effective queries.
        - `global_semantic_search_entity`: Perform semantic search across all data fields to find meaningful matches.
        - `targeted_semantic_search_entity`: Perform semantic search on a specific data field to find meaningful matches.

        Instructions:
        1. Building a plan:
        - First ALWAYS analyse the user question and available tools and build a proper plan.
        - ALWAYS break down the user question logically before acting.
        - You can **revise** , **modify** or **update** the plan.
        2. Then ALWAYS use `get_tables` to see what tables are available.
        3. ALWAYS use `get_user_guidance` to retrive any business-defined logic, get sample query example and logic definitions.
        4. ALWAYS use `get_schema` to explore table structure before querying.
        5. Text Formatting: For 'ALPHANUMERIC', 'CLOSE_ENDED_TEXT', and 'OPEN_ENDED_TEXT' columns, convert all query values (in WHERE clauses and selected columns) to lowercase and remove leading/trailing whitespace. **No trim** and **case standardisation** for columns of datatype "INTEGER" , "FLOAT" and columns involved in SQL JOIN's.
        6. Use `execute_query` for executing queries.
        7. Correct the query by ensuring function compatibility with the {dialect}, fixing syntax issues, handling edge cases (e.g., NULLs, division by zero, case sensitivity, whitespace), and re-testing using the `execute_query` tool.
        8. Use either `global_semantic_search_entity` or `targeted_semantic_search_entity` to resolve entity ambiguity or empty results by performing semantic search on shortlisted fields, falling back to pattern or logic-based queries for others. 
        9. Use `targeted_semantic_search_entity` if you want to do a semantic search within a specific column of a table, use `global_semantic_search_entity` if not sure about the column within which you want to do a search.
        10. Handle errors gracefully:
        - Explain failures.
        - Suggest corrections or alternatives.
        11. Proactive Clarification and Disambiguation:
        Pause and request clarification from the user only after you have analyzed all relevant tables and table schemas and checked for user guidance, and still:
            - The query remains ambiguous, incomplete, or uses undefined terms, or
            - The result of execute_query has multiple possible interpretations.    
            When asking for clarification, always ground your question in findings obtained through appropriate tools to ensure the clarification is relevant and well-informed.

        {universal_instructions}

        NOTE: **Think step-by-step and communicate clearly and concisely**
        """)  # noqa: E501

        if len(universal_instructions.strip()) != 0:
            universal_instructions = f"""\n\n## Universal Instructions (MUST FOLLOW): \n{universal_instructions}"""

        dialect_prompts = {
            SQLDialect.SQLITE: prompt.format(
                source=dialect,
                dialect=dialect.prompt_dialect,
                domain=domain,
                universal_instructions=universal_instructions,
            ),
            SQLDialect.POSTGRES: prompt.format(
                source=dialect,
                dialect=dialect.prompt_dialect,
                domain=domain,
                universal_instructions=universal_instructions,
            ),
            SQLDialect.MSSQL: prompt.format(
                source=dialect,
                dialect=dialect.prompt_dialect,
                domain=domain,
                universal_instructions=universal_instructions,
            ),
        }
        return dialect_prompts[dialect]



from intugle.common.resources.base import BaseResource
from intugle.models.resources.model import Model
from intugle.models.resources.relationship import Relationship
from intugle.models.resources.source import Source


class Resource:
    model_factory = {
        'models': Model,
        'relationships': Relationship,
        'sources': Source
    }

    def create_resource(self, resouce_type: str, data: dict):
        model_cls = self.model_factory.get(resouce_type.lower())
        if not model_cls:
            raise ValueError(f"Resource model: {resouce_type}")
        return model_cls(**data)
    
    @classmethod
    def get_resource(self, resource_type: str) -> BaseResource:
        return self.model_factory.get(resource_type)



from intugle.core.settings import settings
from intugle.mcp.adapter.psql import PsqlPool

# from intugle.parser.manifest import ManifestLoader


class AdapterService:
    """
    Adapter service for executing queries.
    """

    def __init__(self):
        # self.manifest = ManifestLoader.manifest
        self.psql = PsqlPool(
            user=settings.POSTGRES_USER,
            password=settings.POSTGRES_PASSWORD,
            host=settings.POSTGRES_HOST,
            database=settings.POSTGRES_DB,
            port=settings.POSTGRES_PORT,
            schema=settings.POSTGRES_SCHEMA,
        )

    async def execute_query(self, sql_query: str) -> list[dict]:
        """
        Execute a SQL query and return the result.

        Args:
            sql_query (str): The SQL query to execute.

        Returns:
            list[dict]: The result of the query execution.
        """

        data = await self.psql.fetch(sql_query)

        data = [dict(record) for record in data] if data else []

        return data


adapter_service = AdapterService()


from mcp.server.fastmcp import FastMCP

from intugle.core.settings import settings
from intugle.mcp.adapter.service import adapter_service

adapter_mcp = FastMCP(
    name=settings.MCP_SERVER_NAME,
    stateless_http=settings.MCP_SERVER_STATELESS_HTTP,
)


@adapter_mcp.tool(name="execute_query", description="Return the result of a query execution")
async def execute_query(sql_query: str) -> list[dict]: 
    data = adapter_service.execute_query(sql_query)
    print(data)
    return data



from abc import ABC, abstractmethod
from typing import Any

from intugle.adapters.models import (
    ColumnProfile,
    ProfilingOutput,
)


class Adapter(ABC):
    @abstractmethod
    def profile(self, data: Any) -> ProfilingOutput:
        pass

    @abstractmethod
    def column_profile(
        self,
        data: Any,
        table_name: str,
        column_name: str,
        total_count: int,
        sample_limit: int = 10,
        dtype_sample_limit: int = 10000,
    ) -> ColumnProfile:
        pass
    
    @abstractmethod
    def load():
        ...

    @abstractmethod
    def execute():
        raise NotImplementedError()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Any


class BaseExceptionMixin(Exception):
    """Basic exception mixin class"""

    code: int

    def __init__(self, msg: str = None, *, data: Any = None):
        self.msg = msg
        self.data = data


class CustomError(BaseExceptionMixin):
    """Custom Exception"""

    def __init__(self, msg: str, *, code: int, data: Any = None):
        self.code = code
        super().__init__(msg=msg, data=data)


class UnprocessableEntityError(BaseExceptionMixin):
    """Unprocessable Entity"""

    code = 422

    def __init__(self, msg: str = "Unprocessable Entity", *, data: Any = None):
        super().__init__(msg=msg, data=data)


class NotFoundError(BaseExceptionMixin):
    """Not Found"""

    code = 404

    def __init__(self, msg: str = "Not found", *, data: Any = None):
        super().__init__(msg=msg, data=data)


class RuntimeError(BaseExceptionMixin):
    """Runtime Error"""
    
    code = 500

    def __init__(self, msg: str = "Runtime error", *, file: str = None, data: str = None):
        super().__init__(msg=msg, data=data)
        self.file = file


class ParseError(RuntimeError):
    """Parse Error"""

    code = 400



from typing import List

from pydantic import BaseModel

from intugle.common.exception import errors
from intugle.models.resources.relationship import Relationship, RelationshipTable, RelationshipType


class PredictedLink(BaseModel):
    """
    Represents a single predicted link between two columns from different datasets.
    """

    from_dataset: str
    from_column: str
    to_dataset: str
    to_column: str

    @property
    def relationship(self) -> Relationship:
        source = RelationshipTable(table=self.from_dataset, column=self.from_column)
        target = RelationshipTable(table=self.to_dataset, column=self.to_column)
        relationship = Relationship(
            name=f"{self.from_dataset}-{self.to_dataset}",
            description="",
            source=source,
            target=target,
            type=RelationshipType.ONE_TO_MANY,
        )
        return relationship
    

class LinkPredictionResult(BaseModel):
    """
    The final output of the link prediction process, containing all discovered links.
    """

    links: List[PredictedLink]

    @property
    def relationships(self) -> list[Relationship]:
        relationships: list[Relationship] = []
        for link in self.links:
            source = RelationshipTable(table=link.from_dataset, column=link.from_column)
            target = RelationshipTable(table=link.to_dataset, column=link.to_column)
            relationship = Relationship(
                name=f"{link.from_dataset}-{link.to_dataset}",
                description="",
                source=source,
                target=target,
                type=RelationshipType.ONE_TO_MANY,
            )

            relationships.append(relationship)

        return relationships

    def graph(self):
        if len(self.relationships) <= 0:
            raise errors.NotFoundError("No relationships found")

        ...



from intugle.common.exception import errors

# from intugle.mcp.manifest import manifest_loader
from intugle.core import settings
from intugle.parser.manifest import ManifestLoader
from intugle.parser.table_schema import TableSchema


class SemanticLayerService:
    def __init__(self):
        manifest_loader = ManifestLoader(settings.PROJECT_BASE)
        manifest_loader.load()
        self.manifest = manifest_loader.manifest

        self.table_schema = TableSchema(self.manifest)

    async def get_tables(self) -> list[dict]:
        """
        Fetches all the tables and their technical description for a subscription.

        Returns:
            list[dict]: List of tables along with their technical description.
        """
        sources = self.manifest.sources

        tables = []
        for source in sources.values():
            table_info = {
                "table_name": source.table.name,
                "table_description": source.table.description,
            }
            tables.append(table_info)

        return tables

    async def get_schema(self, tables: list[str]) -> dict[str, str]:
        """
        Fetches the schema along with some sample rows for given database table names.

        Args:
            tables (dict[str, str]): List of table names to fetch the schema.

        Raises:
            Exception: If the table names are not found in the manifest.
        """
        schemas = {}

        for table in tables:
            try:
                schema = self.table_schema.get_table_schema(table)
                schemas[table] = schema
            except errors.NotFoundError:
                ...
        
        return schemas


semantic_layer_service = SemanticLayerService()



from pydantic import Field

from intugle.common.schema import SchemaBase
from intugle.models.resources.model import Model
from intugle.models.resources.relationship import Relationship
from intugle.models.resources.source import Source


class Manifest(SchemaBase):
    sources: dict[str, Source] = Field(default_factory=dict)
    models: dict[str, Model] = Field(default_factory=dict)
    relationships: dict[str, Relationship] = Field(default_factory=dict)



import os

from pathlib import Path

import yaml

ALIAS = {
    "DI": "datatype_identification",
    "LP": "link_prediction",
    "KI": "key_identification",
    "BG": "business_glossary",
    "DC": "domain_classification"
    }

ALLOWED_PIPELINE_NAME = ["KI", "LP", "BG", "DI", "DC"]

CODE_PATH = Path(os.path.split(os.path.abspath(__file__))[0]).parent


def load_model_configuration(pipeline_name: str, custom_config: dict):
    '''
        For loading each pipeline configuration
    '''

    assert pipeline_name in ALLOWED_PIPELINE_NAME, f"[!] pipeline name can be only {ALLOWED_PIPELINE_NAME} not {pipeline_name}"
    
    default_config = {}

    if pipeline_name not in ["DC"]:
        default_config_file = os.path.join(CODE_PATH, "pipeline", ALIAS[pipeline_name.upper()], "config.yaml")
        with open(default_config_file, "r") as d_config_file:
            default_config = yaml.safe_load(d_config_file)
            
    if pipeline_name in ["DC"]:
        default_config = {
            
            "INCLUDE_DTYPES": True
        }
    try:
        user_config = custom_config[pipeline_name.upper()] 
    except Exception:
        user_config = {}

    # update the default config if there was a custom configuration
            
    for key in user_config.keys():
        if key in default_config.keys(): 
            if isinstance(default_config[key], dict):
                if len(default_config[key]) <= 0:  # add new config if there was no config under a key 
                    default_config[key] = user_config[key]
                else:  # update the existing config dictionary
                    default_config[key].update(user_config[key])
            else:
                default_config[key] = user_config[key]
                
        else:  # if the corresponding config key not found then add a new config under that keys
            default_config[key] = user_config[key]
    
    return {**default_config}
    


import json
import logging

import pandas as pd

log = logging.getLogger(__name__)

DTYPE_MAPPING = {
    "integer": "INTEGER",
    "float": "FLOAT",
    "date & time": "DATETIME",
    "close_ended_text": "TEXT",
    "open_ended_text": "TEXT",
    "alphanumeric": "TEXT",
    "others": "TEXT",
    "range_type": "TEXT"
}


def generate_create_table_query(table_columns: list,
                                   column_datatypes: dict = {},
                                   table_name: str = "",
                                   profiling_data: pd.DataFrame = None,
                                   primary_keys: list = [],
                                   columns_required: list = [],
                                   mapping_dtypes_to_sql: bool = False
                                   ):
    
    create_table_query = f"CREATE TABLE {table_name.lower()} ("
    column_parts = []
    lngth = len(table_columns)

    for i, column in enumerate(table_columns, 1):
        parts = []
        parts.append(f"{column}")
        column_id = (table_name, column)    
        if column_id in column_datatypes:
            datatype = column_datatypes[column_id] if not mapping_dtypes_to_sql else DTYPE_MAPPING[column_datatypes[column_id]]
            datatype = "(" + datatype + ")"
        else:
            datatype = "VARCHAR(20)"

        parts.append(f"{datatype}")

        pk = ""
        if column_id in primary_keys:
            pk = "PRIMARY KEY"
        
        pk = f"{pk},".lstrip()
        if i == lngth:
            pk = pk.rstrip(",")

        parts.append(pk)

        profiling = ""
        if profiling_data is not None:
            
            profiling = profiling_data.loc[
                (profiling_data["column_name"] == column) & 
                (profiling_data["table_name"] == table_name), 
                list(columns_required)
            ].to_dict(orient="records")[0]

            sample_data = ""

            profiling = json.dumps(profiling).replace("{", "").replace("}", "").replace('"', '').replace("_l1", "")
            profiling = f" -- {profiling}{sample_data}"
        
        parts.append(profiling)
        column_parts.append(" ".join(parts))

    create_table_query = "\n".join([create_table_query] + column_parts + [");"])
    return create_table_query


def read_column_datatypes(dtype: pd.DataFrame,):
    column_datatypes = {}
    try:
        for row in range(len(dtype)): 
            table_name = dtype['table_name'][row].lower()
            column_name = dtype['column_name'][row].lower()
            datatype = dtype['datatype_l1'][row]
            column_datatypes[(table_name, column_name)] = datatype
    except Exception as ex:
        log.warning(f"[!] Error while getting column datatypes: {ex}")
        
    return column_datatypes









from copy import deepcopy

from ..models.models import (
    CategoryType,
    ETLModel,
    FieldDetailsModel,
    FieldType,
    RangeOptions,
)

INITIAL_FILTERS = {"fields": [], "condition": "AND"}


class GroupBy:
    def __init__(
        self,
        etl: ETLModel,
        field_details: dict[int, FieldDetailsModel],
        data_categories: dict[str, list],
    ):
        self.__keys = []
        self.__having = deepcopy(INITIAL_FILTERS)
        self.__etl = etl
        self.__field_details = field_details
        self.__data_categories = data_categories

    def build_keys(self):
        is_mixed_category = False
        _keys = []
        if (
            len(self.__data_categories["dimension"]) > 0
            and len(self.__data_categories["measure"]) > 0
        ):
            is_mixed_category = True

        if is_mixed_category:
            for _field in self.__etl.fields:
                if (
                    _field.id in self.__data_categories["dimension"]
                    and _field.category == CategoryType.dimension
                ):
                    field_detail = self.__field_details[_field.id]
                    renameCol = (
                        _field.name
                        if field_detail.type == FieldType.calculated_field or _field.dimension_func is not None
                        else None
                    )
                    _keys.append(
                        {
                            "dataset": field_detail.asset_name,
                            "field": field_detail.name,
                            "sql_code": field_detail.sql_code,
                            "renameCol": renameCol,
                        }
                    )
            self.__keys = _keys

    def __get_field_name(self, id):
        for _field in self.__etl.fields:
            if _field.id == id:
                return _field.name

    def build_having(self):
        fields = []
        measure_filter = self.__etl.measure_filter
        if measure_filter is None:
            return
        for range in measure_filter.range:
            if range.option == RangeOptions.none:
                continue
            # field_name = self.__get_field_name(range.id)
            field_name = range.alias
            if field_name is None:
                continue
            if range.option != RangeOptions.at_most:
                fields.append(
                    {
                        "sql_code": field_name,
                        "operator": "greater_than_equal_to",
                        "params": {"value": range.value[0]},
                    }
                )
            if range.option != RangeOptions.at_least:
                fields.append(
                    {
                        "sql_code": field_name,
                        "operator": "less_than_equal_to",
                        "params": {"value": range.value[1]},
                    }
                )

        self.__having["fields"].extend(fields)

    def build(self):
        self.build_keys()
        self.build_having()

    def get_group_by(self):
        return {"keys": self.__keys, "having": self.__having}



import contextlib

from logging import getLogger

import uvicorn

from starlette.applications import Starlette
from starlette.routing import Mount

from intugle.core.settings import settings
from intugle.mcp.adapter.router import adapter_mcp
from intugle.mcp.semantic_layer.router import semantic_layer_mcp

log = getLogger(__name__)

# Mount the semantic layer MCP server
# semantic_layer_mcp.settings.mount_path = "/semantic_layer"


# Create a combined lifespan to manage both session managers
@contextlib.asynccontextmanager
async def lifespan(app: Starlette):
    async with contextlib.AsyncExitStack() as stack:
        await stack.enter_async_context(semantic_layer_mcp.session_manager.run())
        yield


# Create Starlette app with multiple mounted servers
app = Starlette(
    routes=[
        # Using settings-based configuration
        Mount("/semantic_layer", app=semantic_layer_mcp.streamable_http_app()),
        Mount("/adapter", app=adapter_mcp.streamable_http_app()),
    ],
    lifespan=lifespan
)


if __name__ == "__main__":
    log.info(f"[*] MCP server started at {settings.MCP_SERVER_HOST}:{settings.MCP_SERVER_PORT}")
    uvicorn.run(
        app,
        host=settings.MCP_SERVER_HOST,
        port=settings.MCP_SERVER_PORT,
        log_level=settings.MCP_SERVER_LOG_LEVEL,
        # reload=True,
    )



from typing import Any, List

from .models import DataSet
from .steps import AnalysisStep


class Pipeline:
    def __init__(self, steps: List[AnalysisStep]):
        self.steps = steps

    def run(self, df: Any, name: str) -> DataSet:
        """
        Executes all analysis steps in order and returns the final dataset.
        """
        dataset = DataSet(df, name)
        for step in self.steps:
            step.analyze(dataset)
        return dataset



# ====================================================================
# mode: To compute the mode for the given data
# Parameters: 
#     axis - Computation based on row or column level axis
#     pre_sorted - sort the data or not
# ====================================================================

def mode(axis, pre_sorted: bool = False):
    if not pre_sorted:
        axis = sorted(axis)

    _count_max = 1
    _count = 0
    _mode = _current = axis[0]

    for v in axis:
        if v == _current:
            _count = _count + 1
        else:
            if _count > _count_max:
                _count_max = _count
                _mode = _current
            _count = 1
            _current = v

    if _count > _count_max:
        return _current
    
    return _mode



import math

from collections import defaultdict
from functools import cmp_to_key

import networkx as nx

from ..models.models import JoinOpt, LinkModel


class Join:
    def __init__(self, links: list[LinkModel], selected_fields: list[str]):
        self._links = links
        self.selected_fields = set(selected_fields)

    def check_if_selected(self, link: LinkModel):
        if (
            link.source_field_id in self.selected_fields
            or link.target_field_id in self.selected_fields
        ):
            return True
        return False

    def get_link_between_src_target(self, source, target):
        connected_asset = None
        fields_selected = False
        res = []
        for link in self._links:
            source_asset_id_in_link = link.source_asset_id in [source, target]
            target_asset_id_in_link = link.target_asset_id in [source, target]

            if source_asset_id_in_link and target_asset_id_in_link:
                ignore = not self.check_if_selected(link)
                connected_asset = (
                    link.target_asset_id
                    if target == link.target_asset_id
                    else link.source_asset_id
                )
                if not ignore:
                    fields_selected = True
                res.append({**link.model_dump(), "ignore": ignore})

        return res, connected_asset, fields_selected

    def _get_join_cndn(self, nodes_added, current_node):
        for prev_tbl in nodes_added[-1::-1]:
            if prev_tbl == current_node:
                continue
            res, connected_asset, fields_selected = self.get_link_between_src_target(
                current_node, prev_tbl
            )
            if not fields_selected:
                res = list(
                    map(lambda link: {**link, "ignore": False}, res)
                )  # Creates new dictionaries

            if len(res) > 0:
                return res, connected_asset
        return [], None

    @staticmethod
    def check_if_join_exists(graph: nx.MultiGraph, path_request: list[int]):
        graph_sizes = [
            len(c)
            for c in sorted(nx.connected_components(graph), key=len, reverse=True)
        ]

        if len(graph_sizes) == 1:
            return True

        path_request_set = set(path_request)
        path_request_len = len(path_request)
        sub_graphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]
        link_found = False
        for _sub_graph in sub_graphs:
            current_nodes = set(_sub_graph.nodes)
            if len(current_nodes.intersection(path_request_set)) == path_request_len:
                link_found = True
                break

        return link_found

    @staticmethod
    def get_join_type(left, right):
        join_map = {
            (JoinOpt.all, JoinOpt.all): "full",
            (JoinOpt.all, JoinOpt.common): "left",
            (JoinOpt.common, JoinOpt.all): "right",
            (JoinOpt.common, JoinOpt.common): "inner",
        }
        return join_map.get((left, right), None)

    def get_table_node_weights(self, datasets_data: list[int], only_connected: bool = True):
        table_weight = defaultdict(list)
        table_connected_fields = defaultdict(list)
        table_connected_tables = defaultdict(set)

        for link in self._links:
            src_table = link.source_asset_id
            src_field = link.source_field_id
            target_table = link.target_asset_id
            target_field = link.source_field_id

            table_connected_fields[src_table].append(src_field)

            table_connected_fields[target_table].append(target_field)

            table_connected_tables[src_table].add(src_table)

            table_connected_tables[target_table].add(target_table)

        tables = list(table_connected_tables.keys())

        if not only_connected:
            tables.extend(datasets_data)

        for table in set(tables):
            if table in datasets_data:
                table_weight[table] = 0
                continue

            connected_field = table_connected_fields[table]

            set_connected_field = set(connected_field)

            table_weight[table] = len(connected_field) - len(set_connected_field)

        return dict(table_weight)

    @staticmethod
    def plot_graph(graph, path=None, draw_weight=False):
        import matplotlib.pyplot as plt
        import numpy as np

        pos = nx.spring_layout(
            graph, k=0.3 * 1 / np.sqrt(len(graph.nodes())), iterations=2, seed=5
        )

        # Draw the graph (without edge labels initially)
        plt.figure(3, figsize=(20, 20))

        # Draw nodes and edges
        nx.draw(
            graph,
            pos,
            with_labels=True,
            node_color="skyblue",
            node_size=2000,
            font_size=15,
            font_weight="bold",
            edge_color="gray",
        )

        if draw_weight:
            # Manually handle and draw edge labels for multiedges
            edge_labels = {}
            for u, v, key, data in graph.edges(data=True, keys=True):
                label = f"w={data['weight']}"
                if (u, v) not in edge_labels:
                    edge_labels[(u, v)] = []
                edge_labels[(u, v)].append((key, label))

            # Draw edge labels, considering multiple edges
            for (u, v), labels in edge_labels.items():
                for key, label in labels:
                    # For multiple edges, offset the label slightly to avoid overlap
                    x_offset = 0.05 * key  # offset for placing labels of multiple edges
                    y_offset = 0.05 * key
                    label_pos = (
                        (pos[u][0] + pos[v][0]) / 2 + x_offset,
                        (pos[u][1] + pos[v][1]) / 2 + y_offset,
                    )
                    plt.text(
                        label_pos[0],
                        label_pos[1],
                        label,
                        color="red",
                        fontsize=12,
                        ha="center",
                        va="center",
                    )

            # Manually draw node labels for their weights
            node_labels = nx.get_node_attributes(graph, "weight")
            for node, weight in node_labels.items():
                # Position the weight label slightly above the node
                label_pos = pos[node]
                plt.text(
                    label_pos[0],
                    label_pos[1] + 0.03,
                    f"{weight}",
                    fontsize=12,
                    ha="center",
                    va="bottom",
                    color="green",
                )

        if path:
            path_edges = list(zip(path, path[1:]))
            nx.draw_networkx_nodes(graph, pos, nodelist=path, node_color="y")
            nx.draw_networkx_edges(
                graph, pos, edgelist=path_edges, edge_color="y", width=10
            )
            plt.axis("equal")

        # Display the plot
        plt.title("Multigraph with Node and Edge Weights")
        plt.show()

    def generate_graph(self, datasets_data: list[int], only_connected: bool = True):
        graph = nx.MultiGraph()

        tables_weights = self.get_table_node_weights(datasets_data, only_connected)

        for table, weight in tables_weights.items():
            graph.add_node(table, weight=weight)

        # tables = set(sum([[i.source_asset_id, i.target_asset_id] for i in links], []))
        # graph.add_nodes_from(tables)

        for link in self._links:
            # relationship weight
            records_mapped = link.records_mapped
            source_count = link.source_count
            target_count = link.target_count
            if records_mapped == 0:
                records_mapped = 1
            if source_count == 0:
                source_count = 1
            if target_count == 0:
                target_count = 1

            source_mapped = min(link.source_count_distinct / records_mapped, 1)
            target_mapped = min(link.target_count_distinct / records_mapped, 1)

            source_uniqueness = min(link.source_count_distinct / source_count, 1)
            target_uniqueness = min(link.target_count_distinct / target_count, 1)

            # for many to many mapping
            uniqueness_weight = 5 - math.floor(max(source_uniqueness, target_uniqueness) * 4 + 1)

            # for mapping records
            source_mapped_weight = 5 - math.floor(
                source_mapped * 4 + 1
            )  # range from 5 to 1

            target_mapped_weight = 5 - math.floor(
                target_mapped * 4 + 1
            )  # range from 5 to 1

            link_weight = (
                tables_weights[link.target_asset_id]
                + source_mapped_weight
                + target_mapped_weight
                + uniqueness_weight
            )

            link_weight = min(link_weight, 10) + 1

            graph.add_edge(
                link.source_asset_id, link.target_asset_id, weight=link_weight
            )

        # self.plot_graph(graph)

        return graph

    @staticmethod
    def get_shortest_path(graph, source_node, target_nodes):
        _path = None
        _weight = None
        for _node in target_nodes:
            if _node == source_node:
                continue
            try:
                # _p = nx.shortest_path(graph, source=node_added, target=node)
                _p = nx.dijkstra_path(graph, source=_node, target=source_node)
                _wt = nx.dijkstra_path_length(graph, source=_node, target=source_node)
                if (
                    _path is None
                    or len(_path) > len(_p)
                    or _weight is None
                    or _weight < _wt
                ):
                    _path = _p
                    _weight = _wt
            except nx.exception.NetworkXNoPath:
                ...

        if _path is None:
            raise ValueError("Shortest join path not found")
        return _path, _weight

    def get_shortest_path_between_node_group(self, graph, source_nodes, target_nodes):
        _path = None
        _weight = None
        for source_node in source_nodes:
            for target_node in target_nodes:
                if target_node == source_node:
                    continue
                try:
                    # _p = nx.shortest_path(graph, source=node_added, target=node)
                    _p = nx.dijkstra_path(graph, source=source_node, target=target_node)
                    _wt = nx.dijkstra_path_length(
                        graph, source=source_node, target=target_node
                    )
                    if (
                        _path is None
                        or len(_path) > len(_p)
                        or _weight is None
                        or _weight < _wt
                    ):
                        _path = _p
                        _weight = _wt
                except nx.exception.NetworkXNoPath:
                    ...
        if _path is None:
            raise ValueError("Shortest join path not found")
        return _path, _weight

    @staticmethod
    def sort_path_request(datasets_data: list[int], join_opt: dict[str, JoinOpt] = {}):
        return sorted(
            datasets_data,
            key=cmp_to_key(
                lambda item1, item2: 1
                if (
                    join_opt.get(str(item1), JoinOpt.common) != JoinOpt.all
                    and join_opt.get(str(item2), JoinOpt.common) == JoinOpt.all
                )
                else -1
            ),
        )

    def get_join_json(
        self, datasets_data: list[int], join_opt: dict[str, JoinOpt] = {}
    ):
        links = self._links

        if len(links) <= 0 and len(datasets_data) <= 1:
            return {"0": {"dataset_id": datasets_data[0]}}
        
        graph = self.generate_graph(datasets_data)

        path_request = self.sort_path_request(datasets_data, join_opt)

        if not self.check_if_join_exists(graph, path_request):
            raise ValueError("Join path not found")

        join = {"0": {"dataset_id": path_request[0]}}

        cur_key = len(join.keys())

        first_table_opt = join_opt.get(str(path_request[0]), JoinOpt.all)

        nodes_added = [path_request[0]]
        remaining_nodes = [*path_request[1:]]

        while len(remaining_nodes) > 0:
            path, _ = self.get_shortest_path_between_node_group(
                graph, nodes_added, remaining_nodes
            )

            for path_node in path:
                if path_node in nodes_added:
                    continue

                fields, _ = self._get_join_cndn(nodes_added, path_node)

                # Used for linking tables
                if path_node not in path_request:
                    _first_table_opt = JoinOpt.all
                else:
                    _first_table_opt = first_table_opt

                second_table_opt = join_opt.get(str(path_node), JoinOpt.common)

                join_type = self.get_join_type(_first_table_opt, second_table_opt)

                join[str(cur_key)] = {
                    "dataset_id": path_node,
                    "join_type": join_type,
                    "fields": fields,
                }

                nodes_added.append(path_node)
                cur_key = cur_key + 1
                
                if path_node in path_request:
                    remaining_nodes.remove(path_node)

                if second_table_opt == JoinOpt.all:
                    first_table_opt = JoinOpt.all

        return join

    @classmethod
    def get_fields(cls, join_json: dict, links: list[LinkModel]):
        fields = set()

        if join_json:
            for join in join_json.values():
                for link in links:
                    if join["dataset_id"] == link.source_asset_id:
                        fields.add(link.source_field_id)
                    elif join["dataset_id"] == link.target_asset_id:
                        fields.add(link.target_field_id)

        return fields



from enum import StrEnum


class SQLDialect(StrEnum):
    POSTGRES = "postgresql"
    MSSQL = "mssql"
    SQLITE = "sqlite"

    def __repr__(self):
        return str(self.value)

    @property
    def sql_glot_dialect(
        self,
    ) -> str:
        if self.value == SQLDialect.MSSQL:
            return "tsql"
        elif self.value == SQLDialect.POSTGRES:
            return "postgres"

        return str(self.value)

    @property
    def prompt_dialect(
        self,
    ) -> str:
        mappings = {
            SQLDialect.MSSQL: "Transact-SQL",
            SQLDialect.POSTGRES: "PostgreSQL",
            SQLDialect.SQLITE: "SQlite",
        }
        return mappings.get(self.value, self.value)



import time

from typing import Any, Optional

import duckdb
import numpy as np
import pandas as pd
import pandas.api.types as ptypes

from intugle.adapters.adapter import Adapter
from intugle.adapters.factory import AdapterFactory
from intugle.adapters.models import (
    ColumnProfile,
    ProfilingOutput,
)
from intugle.adapters.types.duckdb.models import DuckdbConfig
from intugle.adapters.types.pandas.utils import convert_to_native
from intugle.common.exception import errors
from intugle.core.utilities.processing import string_standardization


class DuckdbAdapter(Adapter):
    def profile(self, data: DuckdbConfig) -> ProfilingOutput:
        """
        Generates a profile of a file.

        Args:
            df: The input pandas DataFrame.

        Returns:
            A pydantic model containing the profile information:
            - "count": Total number of rows.
            - "columns": List of column names.
            - "dtypes": A dictionary mapping column names to generalized data types.
        """
        if not isinstance(data, DuckdbConfig):
            raise TypeError("Input must be a duckdb config.")

        def __format_dtype__(dtype: Any) -> str:
            """Maps dtype to a generalized type string."""
            type_map = {
                "VARCHAR": "string",
                "DATE": "date & time",
                "BIGINT": "integer",
                "DOUBLE": "float",
                "FLOAT": "float",
            }
            return type_map.get(dtype, "string")

        table_name = "__profile_table__"
        self.load(data, table_name)

        # Fetching total count of table
        query = f"""
        SELECT count(*) as count FROM {table_name}
        """
        data = duckdb.execute(query).fetchall()

        total_count = data[0][0]

        # Fetching column name and their data types of table
        query = """
        SELECT column_name, data_type
        FROM duckdb_columns()
        WHERE table_name = ?
        """
        data = duckdb.execute(query, [table_name]).fetchall()

        dtypes = {col: __format_dtype__(dtype) for col, dtype in data}
        columns = [col for col, _ in data]

        return ProfilingOutput(
            count=total_count,
            columns=columns,
            dtypes=dtypes,
        )

    def column_profile(
        self,
        data: DuckdbConfig,
        table_name: str,
        column_name: str,
        total_count: int,
        sample_limit: int = 10,
        dtype_sample_limit: int = 10000,
    ) -> Optional[ColumnProfile]:
        """
        Generates a detailed profile for a single column of a pandas DataFrame.

        It calculates null and distinct counts, and generates two types of samples:
        1.  `sample_data`: A sample of unique values.
        2.  `dtype_sample`: A potentially larger sample combining unique values with
            random non-unique values, intended for data type analysis.

        Args:
            df: The input pandas DataFrame.
            column_name: The name of the column to profile.
            sample_limit: The desired number of items for the data samples.

        Returns:
            A dictionary containing the profile for the column, or None if the
            column does not exist.
        """
        if not isinstance(data, DuckdbConfig):
            raise TypeError("Input must be a duckdb config.")
        
        self.load(data, table_name)

        start_ts = time.time()

        # --- Calculations --- #
        query = f"""
        SELECT 
            COUNT(DISTINCT {column_name}) AS distinct_count,
            SUM(CASE WHEN {column_name} IS NULL THEN 1 ELSE 0 END) AS null_count
        FROM  
            {table_name}
        """
        distinct_null_data = duckdb.execute(query).fetchall()

        distinct_count, null_count = distinct_null_data[0]
        not_null_count = total_count - null_count

        # --- Sampling Logic --- #
        # 1. Get a sample of distinct values.
        sample_query = f"""
        SELECT 
            DISTINCT CAST( {column_name} AS VARCHAR) AS sample_values 
        FROM 
            {table_name} 
        WHERE 
            {column_name} IS NOT NULL LIMIT {dtype_sample_limit}
        """
        data = duckdb.execute(sample_query).fetchall()
        distinct_values = [d[0] for d in data]

        not_null_series = pd.Series(distinct_values)

        if distinct_count > 0:
            distinct_sample_size = min(distinct_count, dtype_sample_limit)
            sample_data = list(np.random.choice(distinct_values, distinct_sample_size, replace=False))
        else:
            sample_data = []

        # 2. Create a combined sample for data type analysis.
        dtype_sample = None
        if distinct_count >= dtype_sample_limit:
            # If we have enough distinct values, that's the best sample.
            dtype_sample = sample_data
        elif distinct_count > 0 and not_null_count > 0:
            # If distinct values are few, supplement them with random non-distinct values.
            remaining_sample_size = dtype_sample_limit - distinct_count

            # Use replace=True in case the number of non-null values is less than the remaining sample size needed.
            additional_samples = list(not_null_series.sample(n=remaining_sample_size, replace=True))

            # Combine the full set of unique values with the additional random samples.
            dtype_sample = list(distinct_values) + additional_samples
        else:
            dtype_sample = []

        # --- Convert numpy types to native Python types for JSON compatibility --- #
        native_sample_data = convert_to_native(sample_data)
        native_dtype_sample = convert_to_native(dtype_sample)

        business_name = string_standardization(column_name)

        # --- Final Profile --- #
        return ColumnProfile(
            column_name=column_name,
            business_name=business_name,
            table_name=table_name,
            null_count=null_count,
            count=total_count,
            distinct_count=distinct_count,
            uniqueness=distinct_count / total_count if total_count > 0 else 0.0,
            completeness=not_null_count / total_count if total_count > 0 else 0.0,
            sample_data=native_sample_data[:sample_limit],
            dtype_sample=native_dtype_sample,
            ts=time.time() - start_ts,
        )

    @staticmethod
    def _get_load_func(data: DuckdbConfig):
        func = {
            "csv": "read_csv",
            "parquet": "read_parquet",
            "xlsx": "read_xlsx",
        }
        ld_func = func.get(data.type)
        if ld_func is None:
            raise errors.NotFoundError(f"Type: {data.type} not supported")

        return f"{ld_func}('{data.path}')"

    def load(self, data: DuckdbConfig, table_name: str):
        ld_func = self._get_load_func(data)

        query = f"""CREATE TABLE IF NOT EXISTS {table_name} AS SELECT * FROM {ld_func};"""

        duckdb.execute(query)

    def execute(): ...


def can_handle_pandas(df: Any) -> bool:
    return isinstance(df, DuckdbConfig)


def register(factory: AdapterFactory):
    factory.register("duckdb", can_handle_pandas, DuckdbAdapter)



# Flag default set to True
_is_first = True


# Function to return the is_first flag
def is_first() -> bool:
    return _is_first


# Function to set the is_first flag
def set_first():
    global _is_first
    _is_first = False


# Function to reset the is_first flag
def reset_first():
    global _is_first
    _is_first = True



from typing import TYPE_CHECKING

from intugle.core import settings
from intugle.libs.smart_query_generator import SmartQueryGenerator
from intugle.libs.smart_query_generator.models.models import ETLModel, FieldDetailsModel, LinkModel
from intugle.libs.smart_query_generator.utils.join import Join
from intugle.parser.manifest import ManifestLoader

if TYPE_CHECKING:
    from intugle.models.resources.model import Column


class SqlGenerator:
    """Generates SQL queries based on the manifest and ETL configurations."""

    def __init__(self, project_base: str = settings.PROJECT_BASE):
        self.manifest_loader = ManifestLoader(project_base)
        self.manifest_loader.load()
        self.manifest = self.manifest_loader.manifest

        self.project_base = project_base

        self.field_details = self.get_all_field_details()

        # get the links from the manifest
        self.links = self.get_links()

        selected_fields = set(self.field_details.keys())
        self.join = Join(self.links, selected_fields)

    def generate_query(self, etl: ETLModel) -> str:
        """Generates a SQL query based on the ETL model.

        Args:
            etl (ETLModel): The ETL model containing the configuration for the query.

        Returns:
            str: The generated SQL query.
        """
        # get the field details fetcher function (Not required now as we are getting all field details)
        # and all field details
        field_details_fetcher = self.get_field_details_fetcher()

        # Generate the SQL query using the SmartQueryGenerator
        sql_builder = SmartQueryGenerator(etl, field_details_fetcher, self.links, self.field_details)
        sql_builder.build()
        sql_query = sql_builder.get_query()

        return sql_query

    def get_all_field_details(self) -> dict[str, FieldDetailsModel]:
        """Fetches all field details from the manifest."""

        # get sources from the manifest
        sources = self.manifest.sources

        field_details: dict[str, FieldDetailsModel] = {}

        # iterate through each source and get the field details (all fields / columns)
        for source in sources.values():
            for column in source.table.columns:
                field_detail: FieldDetailsModel = FieldDetailsModel(
                    id=f"{source.table.name}.{column.name}",
                    name=column.name,
                    datatype_l1=column.type,
                    datatype_l2=column.category,
                    sql_code=f"{source.table.name}.{column.name}",
                    is_pii=False,
                    asset_id=source.table.name,
                    asset_name=source.table.name,
                    asset_details={},
                    connection_id=source.schema,
                    connection_source_name="postgresql",
                    connection_credentials={},
                )
                field_details[field_detail.id] = field_detail

        return field_details

    def get_field_details_fetcher(self):
        """Returns field details fetcher from the manifest."""

        manifest = self.manifest

        def field_details_fetcher(ids: list[str]):
            table_columns: dict[str, list[str]] = {}

            # get table and column names from the ids
            for field in ids:
                table, column = field.split(".")
                if table not in table_columns:
                    table_columns[table] = []
                table_columns[table].append(column)

            sources = manifest.sources

            field_details: dict[str, FieldDetailsModel] = {}

            # iterate through each table and column to get the field details
            for table, columns in table_columns.items():
                table_detail = sources.get(table)
                if table_detail is None:
                    continue

                column_details: dict[str, Column] = {}
                for column in table_detail.table.columns:
                    column_details[column.name] = column

                for column in columns:
                    column_detail = column_details[column]
                    field_detail: FieldDetailsModel = FieldDetailsModel(
                        id=f"{table}.{column}",
                        name=column_detail.name,
                        datatype_l1=column_detail.type,
                        datatype_l2=column_detail.category,
                        sql_code=f"{table}.{column}",
                        is_pii=False,
                        asset_id=table,
                        asset_name=table,
                        asset_details={},
                        connection_id=table_detail.schema,
                        connection_source_name="postgresql",
                        connection_credentials={},
                    )
                    field_details[field_detail.id] = field_detail

            return field_details

        return field_details_fetcher

    def get_links(self) -> list[LinkModel]:
        """Fetches the links from the manifest."""

        # get relationships from the manifest
        relationships = self.manifest.relationships
        links: list[LinkModel] = []

        # iterate through each relationship and create a LinkModel
        for relationship in relationships.values():
            links.append(relationship.link)
        return links

    def plot_graph(self, graph):
        self.join.plot_graph(graph)

    def plot_sources_graph(self):
        assets = {field.asset_id for field in self.field_details.values()}

        graph = self.join.generate_graph(list(assets), only_connected=False)

        self.plot_graph(graph)



from langchain.output_parsers import ResponseSchema

table_glossary = [ResponseSchema(name="table_glossary", description=" single-sentence business glossary definition")]
column_glossary = [ResponseSchema(name="column_glossary", description="precise, single-sentence and non-technical business glossary definition")]
column_tag_glossary = [ResponseSchema(name="column_tag_glossary", description="three precise and distinct business tags", type="list[str]")]

BUSINESS_GLOSSARY_PROMPTS = {
"gpt-4o": {
"TABLE_GLOSSARY_TEMPLATE": """You are responsible for Data Governance in {domain},
generate a concise, non-technical business glossary definition for the table on a provided DDL statement.
The definition should be written as a single sentence and clearly describe the business purpose or function.\n
# Instructions
- **Interpret the table name**: Provide a one-sentence, non-technical summary of the table's role from a business perspective (e.g., what type of business entity or process it represents).
- **Input format**: The input will be a DDL statement defining the schema of a table, including table name, columns, and data types. Comments prefixed by `--` may provide contextual hints.
- **Output format**: DONOT mention table name in the glossary description. DONOT start sentence with 'this table'.\n
- **Additional Context**: if provided, `Additional Context` may provide more information about the data.

# Input
{create_statements}
\n
# Additional Context:
{additional_context}\n
# Output
{format_instructions}    
""",
"BUSINESS_GLOSSARY_TEMPLATE": """You are responsible for Data Governance in {domain},
generate a concise single-sentence business glossary definition for each column mentioned in the DDL statement.\n
The definition should clearly describe the business purpose or function.\n

# Instructions:
1.Parse the DDL statement defining the schema, which includes the table name, column names, data types, and any comments (prefixed by --), which may provide additional context about the column.
2.Provide a single-sentence, non-technical business glossary definition for each column, describing what the column represents and how it is used within the business context.
3.If the comment or the name of the column suggests a specific business process, use that context to shape the definition.
4.Don't mention data values in description or any statistics in description.\n\n
5.If provided `Additional Context` may provide more information about the data.

# Input
{create_statements}\n
{format_instructions}
""",
"BUSINESS_TAGS_TEMPLATE": """You are responsible for Data Governance in {domain}, your task is to generate three business tags for a column based on the DDL statements of a table given below.
Use the column's context within the DDL statement (e.g., its name, type, and table name) to infer relevant business tags. Focus on generating concise, domain-relevant, 
and meaningful tags that align with the potential business use of the column.

# Instructions
1. **Parsing the DDL Statement:**  
- Consider the table name and other column definitions ( like sample data) to infer contextual meaning.
- Don't mention data values in description or any statistics in description.

2. **Tag Principles:**  
- Tags should be descriptive of the column's business context, not just technical.
- Tags can represent the purpose, or connection to broader business use.

3. **Number of Tags:**  
- Always generate exactly three distinct business tags.

4. **Language and Style:**  
- Prefer title case consistently across all tags.

5. **Additional Context**
- If provided `Additional Context` may provide more information about the data.

# Input
{create_statements}\n
# Additional Context:
{additional_context}\n
{format_instructions}
"""
},
"gpt-4o-mini": {
"TABLE_GLOSSARY_TEMPLATE": """
Role: You are responsible for Data Governance in the {domain}.\n
Task: You will be given a SQL DDL statement how `{table}` table is structured. Generate a concise, non-technical business glossary definition for `{table}` that clearly describe the business purpose or function.\n

Instructions:\n
- Parse and interpret SQL DDL statement, including attribute names. Comments prefixed but `--` may provide contextual hints.
- Provide a single-sentence, non-technical summary of the table's role from a business perspective (e.g., what type of business entity or process it represents).
- If the comment or the name of the attribute name , or name of the table suggests a specific business process, use that context to shape the definition.
- The glossary should begin with defintion followed by the business role.
- DONOT mention table name in the glossary. 
- DONOT start sentence with either `This table` or `This entity`, or with noun phrase or demonstrative pronoun or `Represents`.
- DONOT mention {domain} in the glossary.
- If provided, `Additional Context` may provide more information about the data.

#Input:\n
{create_statements}\n
# Additional Context:
{additional_context}
\n\n
{format_instructions}
""",
"BUSINESS_GLOSSARY_TEMPLATE": """
Role: You are responsible for Data Governance in the {domain}.\n
Task: You will be given a SQL DDL statement how the attribute `{column}` is structured.\n

Instruction:
1. Interpret the attribute name `{column}`, i.e what type of business entity or process it represents.
2. Parse and interpret SQL DDL statement, including attribute `{column}`. Comments prefixed but `--` may provide contextual hints.
3. Provide a precise single-sentence, non-technical glosary of the attribute `{column}` role from a business perspective.
4. If the comment or the name of the attribute or name of the table suggests a specific business process, use that context to shape the definition.
5. DONOT mention any data values.
6. DONOT include any assumptions.
7. DONOT include attribute name `{column}` in the glossary.
8. DONOT start the glossary with noun phrase or demonstrative pronoun or `This attribute` or `Represents`.
9. DONOT mention {domain} in the glossary.
10. If provided, `Additional Context` may provide more information about the data.\n

#Input:\n
{create_statements}\n
# Additional Context:
{additional_context}\n\n
{format_instructions}
""",
"BUSINESS_TAGS_TEMPLATE":
"""
Role: You are responsible for Data Governance in the {domain}.\n
Task: You will be given a SQL DDL statement how the attribute `{column}` is structured.\n

Instruction:
1. Interpret the attribute name `{column}`, i.e what type of business entity or process it represents.
2. Parse and interpret SQL DDL statement, including attribute `{column}`. Comments prefixed but `--` may provide contextual hints.
3. Focus on generating concise and meaningful tags that align with the potential business use of the column.
4. Tags should be descriptive of the  attribute business context, not just technical.
5. Tags are keywords or labels that can represent the purpose, or connection to broader business use.
6. ALWAYS generate exactly three distinct business tags.
7. Prefer title case consistently across all tags.
8. If provided, `Additional Context` may provide more information about the data.\n

#Input\n
{create_statements}\n
# Additional Context:
{additional_context}\n\n
{format_instructions}
"""
}
}


# ====================================================================
#  Importing the required python packages
# ====================================================================
import logging
import multiprocessing
import os
import random

from collections import OrderedDict
from datetime import datetime

from functional import pseq
from symspellpy.symspellpy import SymSpell

from intugle.core.settings import settings

from .bag_of_words import (
    extract_bag_of_words_features,
)
from .custom_features import extract_addl_feats

# from .paragraph_vectors import (
#     infer_paragraph_embeddings_features,
# )
from .preprocessing import (
    additional_processing,
    special_token_repl,
)

log = logging.getLogger(__name__)


sym_spell = SymSpell(max_dictionary_edit_distance=settings.DI_CONFIG['THRESHOLD']['SYMSPELL_EDIT_DIST'], prefix_length=settings.DI_CONFIG['THRESHOLD']['PARTITION_SIZE'])


dictionary_path = os.path.join(settings.MODEL_DIR_PATH, "dependant", "datatype_l1_identification", "en-80k.txt")
                                           
# dictionary_path = os.path.join(container_path,model_dep_path,'en-80k.txt')

# # ====================================================================
# # term_index is the column of the term and count_index is the column of the term frequency
# # ====================================================================
sym_spell.load_dictionary(dictionary_path, term_index=settings.DI_CONFIG['THRESHOLD']['SYMSPELL_TERM_INDEX'], count_index=settings.DI_CONFIG['THRESHOLD']['SYMSPELL_COUNT_INDEX'])


# ====================================================================
# CPU Core count for multiprocessing and size of partition size
# ====================================================================
core_count = multiprocessing.cpu_count()  
size = settings.DI_CONFIG['THRESHOLD']['PARTITION_SIZE']

# ====================================================================
# extract_features: 
#     - Extract the required metadata information from values list before feature creation
#     - Clean the table and column name for the embedding creation
#     - Additional preprocessing on the data and passed to the feature creation
#     - Created features one by one from characters, words, and embedding based features
#     - Created the none related features followed by metadata features added
# Parameters: 
#     col_values - List of data for feature creation
# Returns:
#     Returns the features dictionary for the input data
# ====================================================================


def extract_features(col_values: list) -> OrderedDict:
    # Extract features from raw data
    # reuse_model = True
    
    # Extract the table data column from column values
    table_name = col_values[0]
    column_name = col_values[1]
    col_values = col_values[2:]

    # Cleaning the table name for special token replacement and segmenting the compound words
    table_name_clean = special_token_repl(table_name, suffix='_table_name')
    table_name_clean = sym_spell.word_segmentation(table_name_clean).corrected_string

    # Cleaning the column name for special token replacement and segmenting the compound words
    column_name_clean = special_token_repl(column_name, suffix='_column_name')
    column_name_clean = sym_spell.word_segmentation(column_name_clean).corrected_string

    # Number of samples used for feature creation and total samples present in data
    n_samples = settings.DI_CONFIG['THRESHOLD']['DATA_VALUES_LIMIT']
    n_values = len(col_values)

    features = OrderedDict()

    # Additional processing on the col values with nan and lower case converted
    log.info('Custom Preprocessing started:%s', datetime.now())
    cleaned_population_nan = pseq(map(additional_processing, col_values), processes=core_count, partition_size=size)
    cleaned_population_nan = list(cleaned_population_nan)

    log.info('Custom preprocessing completed:%s', datetime.now())
    uniq_cleaned_population = len(set([val.lower() for val in cleaned_population_nan if len(val) > 0]))

    # Based on the number of values, either sample the data or take the entire values
    if n_samples < n_values:
        random.seed(13)
        cleaned_sample_nan = random.sample(cleaned_population_nan, k=n_samples)
    else:
        n_samples = n_values
        cleaned_sample_nan = cleaned_population_nan

    # Additional processing on the col values without nan and lower case converted
    cleaned_sample_wo_nan = [val for val in cleaned_sample_nan if len(val) > 0]
    cleaned_sample_wo_nan_uncased = [val.lower() for val in cleaned_sample_wo_nan]
    # uniq_cleaned_sample = list(set(cleaned_sample_wo_nan))

    # Extracting the bag of character, words, embedding based features along with additional statistical features
    log.info('=' * 100)
    extract_bag_of_words_features(cleaned_sample_nan, cleaned_sample_wo_nan_uncased, features)
    log.info('=' * 100)
    extract_addl_feats(cleaned_sample_nan, features)
    log.info('*' * 100)

    # Creating the additional info for table level data
    features['table_population'] = n_values
    features['table_sample'] = n_samples
    features['uniq_values_population'] = uniq_cleaned_population

    # Additional new features
    features['uniq_samp_pop_ratio'] = features['uniq_values_sample'] / uniq_cleaned_population if uniq_cleaned_population > 0 else 0
    features['samp_pop_ratio'] = n_samples / n_values if n_values > 0 else 0

    lengths = [len(s) for s in cleaned_population_nan]
    n_none = sum(1 for _l in lengths if _l == 0)

    # Creating features releated to 'NA' and 'None'
    features["none-agg-has_population"] = 1 if n_none > 0 else 0
    features["none-agg-percent_population"] = n_none / n_values if n_values > 0 else 1
    features["none-agg-num_population"] = n_none
    features["none-agg-all_population"] = 1 if n_none == n_values else 0

    # new completeness logic
    features['completeness'] = (n_values - n_none) / n_values

    # Adding the metadata specific information of the features
    features['table_name'] = table_name
    features['column_name'] = column_name
    features['table_name_clean'] = table_name_clean
    features['column_name_clean'] = column_name_clean

    log.info('Completed...')
    log.info('#' * 100)
    return features



import importlib

from typing import Any, Callable

from .adapter import Adapter


class ModuleInterface:
    """Represents a plugin interface. A plugin has a single register function."""

    @staticmethod
    def register() -> None:
        """Register the necessary items in the environment factory."""


def import_module(name: str) -> ModuleInterface:
    """Imports a module given a name."""
    return importlib.import_module(name)  # type: ignore


DEFAULT_PLUGINS = [
    "intugle.adapters.types.pandas.pandas",
    "intugle.adapters.types.duckdb.duckdb",
]


class AdapterFactory:
    dataframe_funcs: dict[str, tuple[Callable[[Any], bool], Callable[..., Adapter]]] = {}

    # LOADER
    def __init__(self, plugins: list[dict] = None):
        if plugins is None:
            plugins = []

        plugins.extend(DEFAULT_PLUGINS)

        for _plugin in plugins:
            plugin = import_module(_plugin)
            plugin.register(self)

    @classmethod
    def register(
        cls,
        env_type: str,
        checker_fn: Callable[[Any], bool],
        creator_fn: Callable[..., Adapter],
    ) -> None:
        """Register a new execution engine type"""
        cls.dataframe_funcs[env_type] = (checker_fn, creator_fn)

    @classmethod
    def unregister(cls, env_type: str) -> None:
        """Unregister a new execution engine type"""
        cls.dataframe_funcs.pop(env_type, None)

    @classmethod
    def create(cls, df: Any) -> Adapter:
        """Create a execution engine type"""
        for checker_fn, creator_fn in cls.dataframe_funcs.values():
            if checker_fn(df):
                return creator_fn()
        raise ValueError(f"No suitable dataframe type found for object of type {type(df)!r}")
